#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish tal expressions.
"""
import re
import types
import json

import zc.resourcelibrary
import zope.tal.taldefs
from zope.interface import implements
from zope.tales.interfaces import ITALESExpression


class JSONEncoder(json.JSONEncoder):

    _slash_re = re.compile('/')

    def encode(self, ob):
        result = json.JSONEncoder.encode(self, ob)
        return self._slash_re.sub('\/', result)

    def default(self, ob):
        if type(ob) == types.GeneratorType:
            return list(ob)
        return json.JSONEncoder.default(self, ob)


class JSONDecoder(json.JSONDecoder):
    pass


class JSONExpression(object):
    implements(ITALESExpression)

    def __init__(self, name, expr, engine):
        self.name = name
        self.expr = expr
        self.compiled = engine.compile(expr)
        self.encoder = JSONEncoder(sort_keys=True, separators=(',', ':'))

    def __call__(self, econtext):
        ob = self.compiled(econtext)
        return self.encoder.encode(ob)

    def __str__(self):
        return '%s expression (%r)' % (self.name, self.expr)

    def __repr__(self):
        return '<%s %s:%s>' % (self.__class__.__name__, self.name, self.expr)


_scriptlocal_exp = re.compile(
    '\s*(?:(?P<var>[a-zA-Z_$][\w$]*)(?:\s|$))(?:\s*(?P<expr>.*))?\s*')


_tal_name_re = re.compile('^%s$' % zope.tal.taldefs.NAME_RE)

class ScriptLocalExpression(object):
    implements(ITALESExpression)

    compiled = None
    exps = None

    script_template = '<script>$.extend(true,ST.local,%s);</script>'

    def __init__(self, name, expr, engine):
        self.exps = filter(None, [s.strip() for s in expr.split(';')])
        self.compile(self.exps, engine)
        self.encoder = JSONEncoder(sort_keys=True, separators=(',', ':'))

    def extract(self, exp, engine):
        match = _scriptlocal_exp.match(exp)
        if match is None:
            raise engine.getCompilerError()(
                ('Invalid script local expression %r'
                 ' (expected: "js_var_name tales_expression").') % exp)
        parts = match.groupdict()
        js_name = parts['var']
        if not js_name:
            raise engine.getCompilerError()(
                'Missing/invalid javascript variable name %r.' % exp)
        tal_exp = parts['expr']
        if not tal_exp:
            if not _tal_name_re.match(js_name):
                raise engine.getCompilerError()(
                    ('When only javascript variable name is specified (%r)'
                     ' it must be a valid TAL variable.') % exp)
            tal_exp = js_name
        return (js_name, tal_exp)

    def compile(self, expressions, engine):
        self.compiled = {}
        for subexp in expressions:
            js_name, tal_exp = self.extract(subexp, engine)
            if js_name in self.compiled:
                raise engine.getCompilerError()(
                    'Duplicate javascript variable name %r in %r.' % (
                        js_name, '; '.join(expressions)))
            self.compiled[js_name] = engine.compile(tal_exp)

    def __call__(self, econtext):
        if not self.compiled:
            return ''
        zc.resourcelibrary.need('schooltool.skin.flourish-scripts')
        vars = {}
        for js_var, exp in self.compiled.items():
            vars[js_var] = exp(econtext)
        return self.script_template % self.encoder.encode(vars)

    def __str__(self):
        return 'scriptlocal expressions (%s)' % list(self.exps)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, list(self.exps))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Security description views for SchoolTool security policy.
"""

from zope.publisher.browser import BrowserView
from zope.component import queryMultiAdapter
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser import absoluteURL

from schooltool.securitypolicy.metaconfigure import getCrowdsUtility
from schooltool.securitypolicy.metaconfigure import getDescriptionUtility
from schooltool.securitypolicy.crowds import getCrowdDescription
from schooltool.securitypolicy.crowds import collectCrowdDescriptions
from schooltool.common.inlinept import InlinePageTemplate

from schooltool.common import SchoolToolMessage as _


class CrowdDescriptionSnippetBase(BrowserView):

    template = None

    def __init__(self, description, view, request, crowd, action, group):
        self.context = description
        self.view = view
        self.request = request
        self.crowd = crowd
        self.action = action
        self.group = group

    def empty(self):
        return False

    def snippets(self):
        if self.empty():
            return []
        content = self.template(
            view=self, context=self.context, request=self.request)
        return [content.strip()]


class CrowdDescriptionSnippets(CrowdDescriptionSnippetBase):

    template = InlinePageTemplate('''
    <p tal:content="view/context/description" />
    ''')

    def empty(self):
        return not self.context.description.strip()


class AggregatedSnippets(CrowdDescriptionSnippetBase):

    def snippets(self):
        descriptions = self.context.getDescriptions()
        result = []
        snippets = [
            queryMultiAdapter(
                (d, self.view, self.request, d.crowd, d.action, d.group),
                default=None)
            for d in descriptions]
        for collection in snippets:
            if collection:
                result.extend(collection.snippets())
        return result


class AccessSettingSnippets(CrowdDescriptionSnippetBase):

    template = ViewPageTemplateFile('templates/access_setting_snippet.pt')

    @property
    def setting(self):
        return self.crowd.settings.getSetting(self.crowd.setting_key)

    @property
    def status(self):
        status = bool(self.setting.getValue())
        return status and _("Enabled") or _("Disabled")


class AdministrationSnippets(CrowdDescriptionSnippetBase):
    """A way to display short crowd description with a link to a legend.

    Not a good way to do that.
    """

    description = _(u'School administration')
    template = InlinePageTemplate('''
    <p>
      <tal:block content="view/description" />
      <a href="#legend_administration">[1]</a>
    </p>
    ''')


class SecurityDescriptions(BrowserView):

    def done_link(self):
        return '%s/settings' % absoluteURL(self.context, self.request)

    def legends(self):
        # XXX: hacky legend builder (for eye-candy)
        factories = getCrowdsUtility().factories
        crowd = factories['administration'](None)
        description = getCrowdDescription(crowd, None, None)
        if description is None:
            return
        collection = queryMultiAdapter(
            (description, None, self.request,
             description.crowd, description.action, description.group),
            default=None)
        if collection is None:
            return
        yield {
            'idx': '1',
            'description': AdministrationSnippets.description,
            'href': 'legend_administration',
            'snippets': collection.snippets(),
            }

    def getDescriptionSnippets(self, description):
        if description is None:
            return []
        collection = queryMultiAdapter(
            (description, self, self.request,
             description.crowd, description.action, description.group),
            default=None)
        if collection is not None:
            return collection.snippets()
        return []

    def getCrowdSnippets(self, action):
        descriptions = collectCrowdDescriptions(action)
        snippets = []
        for description in descriptions:
            snippets.append(self.getDescriptionSnippets(description))
        return snippets

    def getActions(self, group):
        descriptions = getDescriptionUtility()
        action_dict = descriptions.actions_by_group.get(group.__name__, {})
        actions = sorted(action_dict.values(),
                         key=lambda a: '%d %s' % (a.order, a.__name__))
        for action in actions:
            snippets = self.getCrowdSnippets(action)

            yield {
                'action': action,
                'crowds': snippets,
                }

    def getGroups(self):
        util = getDescriptionUtility()
        groups = sorted(util.groups.values(),
                        key=lambda g: g.__name__)
        for group in groups:
            actions = self.getActions(group)
            yield {'group': group, 'actions': list(actions)}

    def update(self):
        self.groups = list(self.getGroups())





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool content provider machinery.
"""
import unittest
import doctest

from zope.component import provideAdapter, provideHandler
from zope.component import queryMultiAdapter
from zope.interface import implements, Interface
from zope.interface.verify import verifyObject

from schooltool.skin.flourish.interfaces import IContentProvider
from schooltool.skin.flourish.interfaces import IContentProviders
from schooltool.skin.flourish.content import ContentProviders
from schooltool.skin.flourish.content import ContentProvider
from schooltool.skin.flourish.content import ContentProviderProxy
from schooltool.app.browser.testing import setUp, tearDown


def doctest_ContentProvider():
    """Tests for ContentProvider base class.

        >>> from zope.contentprovider.interfaces import IBeforeUpdateEvent

        >>> provider = ContentProvider('context', 'request', 'view')
        >>> verifyObject(IContentProvider, provider)
        True

        >>> events = []
        >>> def log_event(event):
        ...     events.append('%s for %s' %
        ...         (event.__class__.__name__,
        ...          event.object.__class__.__name__))
        >>> provideHandler(log_event, [IBeforeUpdateEvent])

        >>> provider()
        Traceback (most recent call last):
        ...
        NotImplementedError: ``render`` method must be implemented by subclass

        >>> events
        ['BeforeUpdateEvent for ContentProvider']

        >>> events[:] = []

        >>> class ProviderForTest(ContentProvider):
        ...     def update(self):
        ...         events.append('%s.update' % self.__class__.__name__)
        ...     def render(self, *args, **kw):
        ...         events.append('%s.render' % self.__class__.__name__)
        ...         return 'rendered with %s %s' % (args, kw)

        >>> provider = ProviderForTest('context', 'request', 'view')

        >>> result = provider('param1', keyword='hello')

        >>> events
        ['BeforeUpdateEvent for ProviderForTest',
         'ProviderForTest.update',
         'ProviderForTest.render']

        >>> print result
        rendered with ('param1',) {'keyword': 'hello'}

        >>> provider.context, provider.request, provider.view
        ('context', 'request', 'view')

    ContentProvider.view is an alternative to ContentProvider.__parent__.

        >>> provider.__parent__ = 'the view'

        >>> provider.view
        'the view'

        >>> provider.view = 'again with the view'

        >>> provider.__parent__
        'again with the view'

    """


def doctest_ContentProviderProxy():
    """Tests for ContentProviderProxy.

        >>> from zope.contentprovider.interfaces import (
        ...     IContentProvider as IZopeContentProvider)
        >>> from zope.contentprovider.interfaces import IBeforeUpdateEvent
        >>> from zope.contentprovider.provider import ContentProviderBase

        >>> class OldProvider(ContentProviderBase):
        ...     def update(self):
        ...         print 'updating', self
        ...     def render(self, *args, **kw):
        ...         print 'rendering', self
        ...         return 'Rendered with arguments: %s; %s' % (args, kw)

    Zope's content providers are not callable:

        >>> old_provider = OldProvider('context', 'request', 'view')

        >>> old_provider()
        Traceback (most recent call last):
        ...
        TypeError: 'OldProvider' object is not callable

    This proxy just wraps zope's content providers, making them callable.
    The proxied provider behaviour of TALESProviderExpression from
    zope.contentprovider.

        >>> provideAdapter(ContentProviderProxy)
        >>> provider = IContentProvider(old_provider)

        >>> def print_event(event):
        ...     print event.__class__.__name__, 'for', event.object
        >>> provideHandler(print_event, [IBeforeUpdateEvent])

        >>> result = provider('arg1', 'arg2', options='some options')
        BeforeUpdateEvent for <...OldProvider object at ...>
        updating <...OldProvider object at ...>
        rendering <...OldProvider object at ...>

        >>> print result
        Rendered with arguments: ('arg1', 'arg2'); {'options': 'some options'}

    The proxied provider is otherwise transparent.

        >>> verifyObject(IZopeContentProvider, provider)
        True

        >>> verifyObject(IContentProvider, provider)
        True

        >>> provider.context, provider.request, provider.__parent__
        ('context', 'request', 'view')

    """


def doctest_ContentProviders_traversal():
    """Tests for ContentProviders.

    ContentProviders are designed as multi adapter to IContentProviders.

        >>> class SomeContext(object):
        ...     pass

        >>> provideAdapter(ContentProviders,
        ...                (SomeContext, None, None),
        ...                IContentProviders)

        >>> def adapt(*args):
        ...     return queryMultiAdapter(args, IContentProviders)

        >>> providers = adapt(SomeContext(), 'request', 'view')

        >>> verifyObject(IContentProviders, providers)
        True

        >>> providers.context, providers.request, providers.view
        (<...SomeContext ...>, 'request', 'view')

   The adapter is traversable - it looks up content providers by
   adapting to IContentProvider.

   Let's register a content provider.

        >>> class ContentProvider(ContentProvider):
        ...     def render(self, *args, **kw):
        ...         return 'Rendered %r for %r' % (self, self.context)

        >>> provideAdapter(ContentProvider,
        ...                (SomeContext, None, None),
        ...                provides=IContentProvider,
        ...                name='dummy')

        >>> dummy = providers.traverse('dummy', ())

        >>> dummy()
        'Rendered <...ContentProvider ...> for <...SomeContext ...>'

    Traversal does not modify further path:

        >>> path = ['more', 'to', 'follow']
        >>> name = 'dummy'
        >>> dummy = providers.traverse(name, path)
        >>> path
        ['more', 'to', 'follow']

    Traversing to non-existent providers throws an exception.

        >>> providers.traverse('info', ())
        Traceback (most recent call last):
        ...
        ContentProviderLookupError: (<...SomeContext ...>, 'info')

    It is possible to traverse to zope's content providers.

        >>> from zope.contentprovider.provider import ContentProviderBase
        >>> class ContentProviderStub(ContentProviderBase):
        ...     def render(self, *args, **kw):
        ...         return 'Rendered %r for %r' % (self, self.context)

        >>> from zope.contentprovider.interfaces import (
        ...     IContentProvider as IZopeContentProvider)
        >>> provideAdapter(ContentProviderStub,
        ...                (SomeContext, None, None),
        ...                provides=IZopeContentProvider,
        ...                name='info')

        >>> providers.traverse('info', ())
        Traceback (most recent call last):
        ...
        ContentProviderLookupError: (<...SomeContext ...>, 'info')

    But they need to be adaptable to IContentProvider.
    Also, content providers cache the looked up providers, so we need to reset
    the cache.

        >>> provideAdapter(ContentProviderProxy)

        >>> info = providers.traverse('info', ())
        Traceback (most recent call last):
        ...
        ContentProviderLookupError: (<...SomeContext ...>, 'info')

        >>> providers = adapt(SomeContext(), 'request', 'view')

        >>> info = providers.traverse('info', ())

        >>> info()
        'Rendered <...ContentProviderStub ...> for <...SomeContext ...>'

    Let's try with another context.

        >>> class OtherContext(object):
        ...     pass

        >>> print adapt(OtherContext(), 'request', 'view')
        None

        >>> provideAdapter(ContentProviders,
        ...                (OtherContext, None, None),
        ...                IContentProviders)

    Content providers are context/request/view sensitive, and we can see
    that there is no 'dummy' ContentProvider for this context.

        >>> providers = adapt(OtherContext(), 'request', 'view')
        >>> providers.traverse('dummy', ())
        Traceback (most recent call last):
        ...
        ContentProviderLookupError: (<...OtherContext ...>, 'dummy')

    """


def doctest_ContentProviders_getitem():
    """Tests demonstrating dict-like ContentProviders behaviour.

    ContentProviders are designed as multi adapter to IContentProviders.

        >>> class SomeContext(object):
        ...     pass

        >>> provideAdapter(ContentProviders,
        ...                (SomeContext, None, None),
        ...                IContentProviders)

        >>> def adapt(*args):
        ...     return queryMultiAdapter(args, IContentProviders)

        >>> providers = adapt(SomeContext(), 'request', 'view')

        >>> class TestProvider(ContentProvider):
        ...     def __repr__(self):
        ...         return '<%s %r>' % (self.__class__.__name__, self.__name__)

        >>> provideAdapter(TestProvider,
        ...                (SomeContext, None, None),
        ...                provides=IContentProvider,
        ...                name='frog')

        >>> provideAdapter(TestProvider,
        ...                (SomeContext, None, None),
        ...                provides=IContentProvider,
        ...                name='pond')

        >>> providers['frog']
        <TestProvider 'frog'>

        >>> providers.get('pond')
        <TestProvider 'pond'>

        >>> providers['spoon']
        Traceback (most recent call last):
        ...
        KeyError: 'spoon'

        >>> providers.get('spoon', default='There is no spoon')
        'There is no spoon'

    """


def doctest_TALESAwareContentProviders():
    """Tests for TALESAwareContentProviders.

    When TALESAwareContentProviders traverse to a content provider,
    they add the requested TAL attributes to the provider.  Tal attributes are
    taken from the tal engine.

        >>> from schooltool.skin.flourish.content import (
        ...     TALESAwareContentProviders)

        >>> providers = TALESAwareContentProviders(
        ...     'context', 'request', 'view')

        >>> class ContentProviderStub(ContentProvider):
        ...     def render(self, *args, **kw):
        ...         return 'Rendered %r for %r' % (self, self.context)

        >>> provideAdapter(ContentProviderStub,
        ...                (None, None, None),
        ...                provides=IContentProvider,
        ...                name='dummy')

        >>> class TALContext(object):
        ...     def __init__(self, vars=()):
        ...         self.vars = dict(vars)

        >>> providers.setEngine(TALContext({
        ...     'hello': 'Hello world!',
        ...     'bar': 'BAR'
        ...     }))

    Simply traversing to a provider will not set the attributes from tal.

        >>> provider = providers.traverse('dummy', ())
        >>> provider.hello
        Traceback (most recent call last):
        ...
        AttributeError: 'ContentProviderStub' object has no attribute 'hello'

    To do this, we need a schema that defines the attributes we want.

        >>> from zope.schema import TextLine

        >>> class IHelloData(Interface):
        ...     hello = TextLine(title=u'The greeting.')

    We also need to set the type of the schema to ITALNamespaceData

        >>> from zope.interface import alsoProvides
        >>> from zope.contentprovider.interfaces import ITALNamespaceData

        >>> alsoProvides(IHelloData, ITALNamespaceData)

   Finally, make our content provider implement IHelloData.

        >>> class GreetingContentProvider(ContentProviderStub):
        ...     implements(IHelloData)

        >>> provideAdapter(GreetingContentProvider,
        ...                (None, None, None),
        ...                provides=IContentProvider,
        ...                name='greeting')

   We can now see that relevant data is passed from tal.

        >>> provider = providers.traverse('greeting', ())

        >>> provider.hello
        'Hello world!'

        >>> provider.bar
        Traceback (most recent call last):
        ...
        AttributeError: 'GreetingContentProvider' object has no attribute 'bar'

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# Make a package.

def registerTestSetup():
    from zope.interface import classImplements
    from schooltool.testing import registry

    def addResourceContainer(app):
        from schooltool.resource import resource
        app['resources'] = resource.ResourceContainer()
    registry.register('ApplicationContainers', addResourceContainer)

    def haveCalendar():
        from schooltool.resource import resource
        from schooltool.app.interfaces import IHaveCalendar
        if not IHaveCalendar.implementedBy(resource.Resource):
            classImplements(resource.Resource, IHaveCalendar)
    registry.register('CalendarComponents', haveCalendar)

registerTestSetup()
del registerTestSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2004 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.browser.timetablecsvimport
"""
import unittest
import doctest
import datetime
import transaction

from zope.app.testing import setup as zope_setup
from zope.interface import Interface
from zope.component import provideAdapter, provideUtility, provideHandler
from zope.component.hooks import getSite, setSite
from zope.i18n import translate
from zope.intid import IntIds
from zope.intid import addIntIdSubscriber
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.location.interfaces import ILocation

from schooltool.app.browser.timetablecsvimport import TimetableCSVImporter
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.person import BasicPerson
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.course import CourseInit
from schooltool.course.course import Course
from schooltool.course.course import getCourseContainer
from schooltool.course.course import getCourseContainerForApp
from schooltool.course.section import SectionInit
from schooltool.course.section import getSectionContainer
from schooltool.course.section import getTermForSection
from schooltool.course.section import getTermForSectionContainer
from schooltool.course.interfaces import ISectionContainer
from schooltool.course.browser.section import SectionNameChooser
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.schoolyear import getSchoolYearContainer
from schooltool.relationship.tests import setUpRelationships
from schooltool.term.interfaces import ITerm
from schooltool.term.term import Term
from schooltool.term.term import getTermContainer
from schooltool.term.term import getSchoolYearForTerm
from schooltool.term.term import listTerms
from schooltool.testing import registry as testing_registry
from schooltool.testing.setup import setUpSchoolToolSite
from schooltool.testing.stubs import KeyReferenceStub
from schooltool.timetable.interfaces import IScheduleContainer
from schooltool.timetable.app import getTimetableContainer
from schooltool.timetable.app import getScheduleContainer
from schooltool.timetable.app import TimetableStartUp
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.daytemplates import CalendarDayTemplates
from schooltool.timetable.daytemplates import WeekDayTemplates
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period
from schooltool.timetable.timetable import Timetable
from schooltool.timetable.browser.tests.test_ttwizard import print_day_templates


def print_schedule(schedule):
    print "Schedule '%s'" % schedule.title
    print 'Periods (%s)' % schedule.periods.__class__.__name__
    print_day_templates(schedule.timetable.periods.templates,
                        filter=lambda p: schedule.hasPeriod(p))


def setUpYear():
    app = ISchoolToolApplication(None)
    syc = ISchoolYearContainer(app)
    syc['2010-2011'] = SchoolYear(u'2010-2011',
                                  datetime.date(2010, 9, 1),
                                  datetime.date(2011, 5, 30))
    syc['2010-2011'][u'fall'] = Term(u'Fall',
                                     datetime.date(2010, 9, 1),
                                     datetime.date(2010, 11, 30))
    syc['2010-2011'][u'winter'] = Term(u'Winter',
                                       datetime.date(2010, 12, 1),
                                       datetime.date(2011, 2, 28))
    syc['2010-2011'][u'spring'] = Term(u'Spring',
                                       datetime.date(2011, 3, 1),
                                       datetime.date(2011, 5, 30))


def setUpPersons(names):
    app = ISchoolToolApplication(None)
    persons = app['persons']
    for name in names:
        parts = name.split(' ')
        first = parts[0]
        last = ' '.join(parts[1:])
        username = parts[-1].lower()
        persons[username] = BasicPerson(username, first, last)


def setUpCourses(names):
    app = ISchoolToolApplication(None)
    CourseInit(app)()
    SectionInit(app)()
    courses = ICourseContainer(app)
    for name in names:
        courses[name.lower()] = Course(title=name)


def addTimetableDays(tt, days, periods):
    for day_id, day_title in days:
        p_days = tt.periods.templates
        day = p_days[day_id] = DayTemplate(day_title)
        for period in periods:
            day[period.lower()] = Period(title=period)

        t_days = tt.time_slots.templates
        day = t_days[day_id] = DayTemplate(day_title)
        for n, period in enumerate(periods):
            time = datetime.time(8+n, 0)
            duration = datetime.timedelta(0, 3300)
            day[period.lower()] = TimeSlot(time, duration)


def setUpTimetables():
    app = ISchoolToolApplication(None)
    TimetableStartUp(app)()
    syc = ISchoolYearContainer(app)
    sy = syc.getActiveSchoolYear()
    timetables = ITimetableContainer(sy)
    timetables[u'rotating'] = tt_rot = Timetable(
        sy.first, sy.last, title=u"Rotating")

    tt_rot.periods = CalendarDayTemplates()
    tt_rot.periods.initTemplates()
    tt_rot.time_slots = CalendarDayTemplates()
    tt_rot.time_slots.initTemplates()
    addTimetableDays(
        tt_rot,
        [('1', 'Day 1'), ('2', 'Day 2'), ('3', 'Day 3')],
        ['A', 'B', 'C'])

    timetables[u'weekly'] = tt_week = Timetable(
        sy.first, sy.last, title=u"Weekly")

    tt_week.periods = WeekDayTemplates()
    tt_week.periods.initTemplates()
    tt_week.time_slots = WeekDayTemplates()
    tt_week.time_slots.initTemplates()

    dows = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
            'Friday', 'Saturday', 'Sunday']
    week_days = [(unicode(n), title)
                 for n, title in enumerate(dows)]

    addTimetableDays(
        tt_week,
        week_days,
        ['A', 'B', 'C'])


person_names = (
'John Black',
'Pete Cook',
'Daniel Lewis',
'Glen Miller',
'Alan Burton',
'Heinrich Lorch',
'Andreas Guzman',
)

course_names = ('Math', 'Philosophy', 'Literature')

simple_csv = """
"Fall","Spring"
""
"philosophy","lorch"
"rotating"
"Day 1","A"
"Day 1","B"
"Day 3","C"
"***"
"cook"
"lewis"
""
"literature","guzman"
"rotating"
"Day 1","B"
"Day 2","C"
"weekly"
"Monday","A"
"***"
"black"
"cook"
""
"math","burton"
"***"
"lewis"
""
""".strip()


def printYearSections(schoolyear, with_ids=False):
    def sectionName(term, section):
        return '%s in %s' % (
            translate(section.label), term.title)
    for term in listTerms(schoolyear):
        sections = ISectionContainer(term)
        for s_name in sorted(sections):
            print '*' * 50
            section = sections[s_name]
            print sectionName(term, section)
            if with_ids:
                print '  ID:', section.__name__
            s_next = section.next
            if s_next is not None:
                n_term = ITerm(s_next)
                print '  next:', sectionName(n_term, s_next)
            print '  students:'
            for member in sorted(section.members, key=lambda m:m.__name__):
                print '   ', member.title
            schedules = IScheduleContainer(section)
            for key in sorted(schedules):
                print_schedule(schedules[key])


def doctest_TimetableCSVImporter():
    """
        >>> setUpYear()
        >>> setUpPersons(person_names)
        >>> setUpCourses(course_names)
        >>> setUpTimetables()

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)
        >>> schoolyear = syc.getActiveSchoolYear()

        >>> importer = TimetableCSVImporter(schoolyear)

        >>> def tryImport(csv):
        ...    result = importer.importFromCSV(csv)
        ...    if not result:
        ...        print importer.errors

        >>> tryImport(simple_csv)

        >>> printYearSections(schoolyear)
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Fall
          next: Heinrich Lorch -- Philosophy (1) in Winter
          students:
            Pete Cook
            Daniel Lewis
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       |       |
        | B     |       |       |
        |       |       | C     |
        +-------+-------+-------+
        **************************************************
        Andreas Guzman -- Literature (2) in Fall
          next: Andreas Guzman -- Literature (2) in Winter
          students:
            John Black
            Pete Cook
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        |       |       |       |
        | B     |       |       |
        |       | C     |       |
        +-------+-------+-------+
        Schedule 'Weekly'
        Periods (list)
        +--------+---------+-----------+----------+--------+----------+--------+
        | Monday | Tuesday | Wednesday | Thursday | Friday | Saturday | Sunday |
        +--------+---------+-----------+----------+--------+----------+--------+
        | A      |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        +--------+---------+-----------+----------+--------+----------+--------+
        **************************************************
        Alan Burton -- Math (3) in Fall
          next: Alan Burton -- Math (3) in Winter
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Winter
          next: Heinrich Lorch -- Philosophy (1) in Spring
          students:
            Pete Cook
            Daniel Lewis
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       |       |
        | B     |       |       |
        |       |       | C     |
        +-------+-------+-------+
        **************************************************
        Andreas Guzman -- Literature (2) in Winter
          next: Andreas Guzman -- Literature (2) in Spring
          students:
            John Black
            Pete Cook
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        |       |       |       |
        | B     |       |       |
        |       | C     |       |
        +-------+-------+-------+
        Schedule 'Weekly'
        Periods (list)
        +--------+---------+-----------+----------+--------+----------+--------+
        | Monday | Tuesday | Wednesday | Thursday | Friday | Saturday | Sunday |
        +--------+---------+-----------+----------+--------+----------+--------+
        | A      |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        +--------+---------+-----------+----------+--------+----------+--------+
        **************************************************
        Alan Burton -- Math (3) in Winter
          next: Alan Burton -- Math (3) in Spring
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Spring
          students:
            Pete Cook
            Daniel Lewis
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       |       |
        | B     |       |       |
        |       |       | C     |
        +-------+-------+-------+
        **************************************************
        Andreas Guzman -- Literature (2) in Spring
          students:
            John Black
            Pete Cook
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        |       |       |       |
        | B     |       |       |
        |       | C     |       |
        +-------+-------+-------+
        Schedule 'Weekly'
        Periods (list)
        +--------+---------+-----------+----------+--------+----------+--------+
        | Monday | Tuesday | Wednesday | Thursday | Friday | Saturday | Sunday |
        +--------+---------+-----------+----------+--------+----------+--------+
        | A      |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        |        |         |           |          |        |          |        |
        +--------+---------+-----------+----------+--------+----------+--------+
        **************************************************
        Alan Burton -- Math (3) in Spring
          students:
            Daniel Lewis


    """


reimport_csv_1 = """
"Winter","Spring"
""
"philosophy","lorch","p-lorch-winter","p-lorch-spring"
"rotating"
"Day 1","A"
"Day 1","B"
"Day 3","C"
"***"
"cook"
"lewis"
""
"math","burton"
"***"
"lewis"
""
""".strip()


reimport_csv_2 = """
"Fall","Spring"
""
"philosophy","lorch","p-lorch-fall","p-lorch-winter","p-lorch-spring"
"rotating"
"Day 1","A"
"Day 2","B"
"Day 3","A"
"***"
"cook"
"miller"
""
"math","burton"
"***"
"lewis"
""
""".strip()


def doctest_TimetableCSVImporter_reimport():
    """
        >>> setUpYear()
        >>> setUpPersons(person_names)
        >>> setUpCourses(course_names)
        >>> setUpTimetables()

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)
        >>> schoolyear = syc.getActiveSchoolYear()

        >>> importer = TimetableCSVImporter(schoolyear)

        >>> def tryImport(csv):
        ...    result = importer.importFromCSV(csv)
        ...    if not result:
        ...        print importer.errors

        >>> tryImport(reimport_csv_1)

    Let's do an initial import, filling only winter and spring terms.

        >>> printYearSections(schoolyear, with_ids=True)
        **************************************************
        Alan Burton -- Math (1) in Winter
          ID: 1
          next: Alan Burton -- Math (1) in Spring
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Winter
          ID: p-lorch-winter
          next: Heinrich Lorch -- Philosophy (1) in Spring
          students:
            Pete Cook
            Daniel Lewis
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       |       |
        | B     |       |       |
        |       |       | C     |
        +-------+-------+-------+
        **************************************************
        Alan Burton -- Math (1) in Spring
          ID: 1
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Spring
          ID: p-lorch-spring
          students:
            Pete Cook
            Daniel Lewis
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       |       |
        | B     |       |       |
        |       |       | C     |
        +-------+-------+-------+

    Let's reimport from an updated csv:

        >>> tryImport(reimport_csv_2)

    And look at what happened:
    - Fall term got new sections.
    - Math lessons taught by Burton got duplicated,
      because no section ids were specified.
    - Philosophy schedule was updated for existing sections.
    - A new student was added to philosophy lessons.
    - No students were removed: reimport only adds new students!
      (seems safer this way)

        >>> printYearSections(schoolyear, with_ids=True)
        **************************************************
        Alan Burton -- Math (1) in Fall
          ID: 1
          next: Alan Burton -- Math (2) in Winter
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Fall
          ID: p-lorch-fall
          next: Heinrich Lorch -- Philosophy (1) in Winter
          students:
            Pete Cook
            Glen Miller
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       | A     |
        |       | B     |       |
        |       |       |       |
        +-------+-------+-------+
        **************************************************
        Alan Burton -- Math (1) in Winter
          ID: 1
          next: Alan Burton -- Math (1) in Spring
          students:
            Daniel Lewis
        **************************************************
        Alan Burton -- Math (2) in Winter
          ID: 2
          next: Alan Burton -- Math (2) in Spring
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Winter
          ID: p-lorch-winter
          next: Heinrich Lorch -- Philosophy (1) in Spring
          students:
            Pete Cook
            Daniel Lewis
            Glen Miller
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       | A     |
        |       | B     |       |
        |       |       |       |
        +-------+-------+-------+
        **************************************************
        Alan Burton -- Math (1) in Spring
          ID: 1
          students:
            Daniel Lewis
        **************************************************
        Alan Burton -- Math (2) in Spring
          ID: 2
          students:
            Daniel Lewis
        **************************************************
        Heinrich Lorch -- Philosophy (1) in Spring
          ID: p-lorch-spring
          students:
            Pete Cook
            Daniel Lewis
            Glen Miller
        Schedule 'Rotating'
        Periods (list)
        +-------+-------+-------+
        | Day 1 | Day 2 | Day 3 |
        +-------+-------+-------+
        | A     |       | A     |
        |       | B     |       |
        |       |       |       |
        +-------+-------+-------+

    """


def setUpIntIds():
    provideAdapter(KeyReferenceStub)
    provideHandler(addIntIdSubscriber,
                   [ILocation, IObjectAddedEvent])
    provideUtility(IntIds(), IIntIds)
    testing_registry.setupTimetablesComponents()


def docSetUp(test=None):
    zope_setup.placefulSetUp()
    zope_setup.setUpAnnotations()
    zope_setup.setUpTraversal()
    test.globs['app'] = setUpSchoolToolSite()
    setUpRelationships()

    provideAdapter(getSchoolYearContainer)
    provideAdapter(getTermContainer, (Interface,))
    provideAdapter(getSchoolYearForTerm)
    provideAdapter(getCourseContainer)
    provideAdapter(getCourseContainerForApp)
    provideAdapter(getSectionContainer)
    provideAdapter(getTermForSection)
    provideAdapter(getTermForSectionContainer)
    provideAdapter(getTimetableContainer)
    provideAdapter(getScheduleContainer)
    setUpIntIds()

    provideAdapter(SectionNameChooser, (ISectionContainer,))

    transaction.begin()


def docTearDown(test=None):
    transaction.abort()
    zope_setup.placefulTearDown()
    setSite()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
        setUp=docSetUp, tearDown=docTearDown,
        optionflags=doctest.ELLIPSIS|doctest.REPORT_NDIFF))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Interfaces for SchoolTool calendar browser views.
"""
import zope.schema
from zope.interface import Interface, Attribute
from zc.table.interfaces import IColumn

from schooltool.skin.flourish.interfaces import IContentProvider


class IBatch(Interface):

    def render():
        """ """


class IFilterWidget(Interface):

    def render():
        """Render the HTML representation of the filtering widget. """

    def filter(list):
        """Process the list leaving only those elements that match the query."""

    def active():
        """Returns True if there is at least one search parameter in the request."""

    def extra_url():
        """String that should be appended to the url to preserve query parameters."""


class ITableFormatter(Interface):

    batch = zope.schema.Object(schema=IBatch)

    filter_widget = zope.schema.Object(schema=IFilterWidget)

    def setUp(items=None, ommit=None, filter=None, columns=None, columns_before=None,
              columns_after=None, sort_on=None, prefix="", formatters=None,
              table_formatter=None, batch_size=10):
        """Populate the table with items, set up the variables for table formatter.

        After calling this method you have batch and filter_widget set
        up as well, so you can render them in the view as well.
        """

    def makeFormatter():
        """Build the zc.table formatter that can render this table."""

    def render():
        """Render the table for display in a view."""


class IIndexedTableFormatter(ITableFormatter):

    def indexItems(items):
        """Return a list of indexed items"""


class IIndexedColumn(Interface):
    """A column that operates on index dicts instead of objects.

    Index dicts are composed this way:

    context - the container containing items
    id      - the int id of the object
    catalog - the catalog that is storing relevant indexes
    key     - the key by which the object can be retrieved from the context

    XXX: information is a bit outdated
    """

class ICheckboxColumn(IColumn):
    """A column with a checkbox."""


class IRMLTable(IContentProvider):
    """Content provider that can render a table as RML."""

    table = Attribute("zc.table.interfaces.IFormatter to render")


class IRMLColumn(IColumn):

    column = Attribute("zc.table.column.Column")

    visible = zope.schema.Bool(
        title=u"Is column visible", required=False)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os.path
import pkg_resources
from paste.script.templates import var, Template

HOME = os.path.expanduser('~')

def get_available_types():
    instance_types = list(pkg_resources.iter_entry_points('schooltool.instance_type'))
    return [(entry.name, entry.module_name) for entry in sorted(instance_types)]

available_types = get_available_types()


def get_paste_parts():
    paste_part_factories = list(pkg_resources.iter_entry_points(
            'schooltool.paste_configuration'))
    return [entry.load() for entry in paste_part_factories]

paste_configurers = get_paste_parts()


class SchoolToolDeploy(Template):
    _template_dir = 'schooltool_template'
    summary = "(Paste) deployment of a SchoolTool application"

    vars = [
        var('instance_type', """SchoolTool instance type to use. Available types -
  %s""" % "\n  ".join([t[0] for t in available_types]),
            default='schooltool')]

    def check_vars(self, vars, cmd):
        vars = super(SchoolToolDeploy, self).check_vars(vars, cmd)
        vars['instance_package'] = dict(available_types)[vars['instance_type']]
        vars['config_dir'] = os.path.join(
            os.path.abspath(cmd.options.output_dir), vars['project'])
        vars['bin_dir'] = os.path.abspath(
            vars.get('bin_dir', 'bin'))
        vars['log_dir'] = os.path.abspath(
            vars.get('log_dir', os.path.join(vars['project'], 'log')))
        vars['data_dir'] = os.path.abspath(
            vars.get('data_dir', os.path.join(vars['project'], 'var')))
        vars['run_dir'] = os.path.abspath(
            vars.get('run_dir', os.path.join(vars['project'], 'run')))
        vars['paste_extra_paths'] = ''
        vars['paste_extra_parts'] = ''
        for factory in paste_configurers:
            factory(vars)
        return vars

    def write_files(self, command, output_dir, vars):
        super(SchoolToolDeploy, self).write_files(command, output_dir, vars)
        for directory in (vars['log_dir'], vars['data_dir'], vars['run_dir']):
            if not os.path.exists(directory):
                print "Creating directory %s" % directory
                if not command.simulate:
                    os.makedirs(directory)



# -*- coding: utf-8 -*-
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for course and section views.
"""
import unittest
import doctest

from zope.i18n import translate
from zope.interface import directlyProvides
from zope.publisher.browser import TestRequest
from zope.publisher.browser import BrowserView
from zope.traversing.interfaces import IContainmentRoot
from zope.component import provideAdapter

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.testing import setup

class AddingStub(BrowserView):
    pass

def doctest_CourseContainerView():
    r"""View for the courses container.

        >>> from schooltool.course.browser.course import CourseContainerView
        >>> from schooltool.course.course import CourseContainer
        >>> cc = CourseContainer()
        >>> request = TestRequest()
        >>> view = CourseContainerView(cc, request)

    """


def doctest_CourseView():
    r"""View for courses.

    Lets create a simple view for a course:

        >>> from schooltool.course.browser.course import CourseView
        >>> from schooltool.course.course import Course
        >>> course = Course(title="Algebra 1")
        >>> request = TestRequest()
        >>> view = CourseView(course, request)

    And look at the course details.

        >>> print view.details
        []

    Add some details and check again.

        >>> course.course_id = u'Course1'
        >>> course.government_id = u'GovC1'
        >>> course.credits = 1
        >>> [sorted(detail.items()) for detail in view.details]
        [[('title', u'Course ID'),    ('value', u'Course1')],
         [('title', u'Alternate ID'), ('value', u'GovC1')],
         [('title', u'Credits'),      ('value', 1)]]

    Empty values are hidden.

        >>> course.course_id = u''
        >>> course.government_id = u'   '
        >>> course.credits = 0
        >>> [sorted(detail.items()) for detail in view.details]
        [[('title', u'Credits'), ('value', 0)]]

    """


def doctest_CourseAddView():
    r"""

        >>> from schooltool.course.browser.course import CourseAddView
        >>> from schooltool.course.course import Course
        >>> from schooltool.course.interfaces import ICourse

        >>> class CourseAddViewForTesting(CourseAddView):
        ...     schema = ICourse
        ...     fieldNames = ('title', 'description')
        ...     _factory = Course

        >>> from schooltool.course.course import CourseContainer
        >>> container = CourseContainer()
        >>> request = TestRequest()
        >>> context = AddingStub(container, request)
        >>> context = container

        >>> view = CourseAddViewForTesting(context, request)
        >>> view.update()

        >>> request = TestRequest()
        >>> request.form = {'field.title' : 'math'}
        >>> context = AddingStub(context, request)
        >>> view = CourseAddViewForTesting(context, request)
        >>> view.update()

    """


def doctest_CourseCSVImporter():
    r"""Tests for CourseCSVImporter.

    Create a course container and an importer

        >>> from schooltool.course.browser.csvimport import CourseCSVImporter
        >>> from schooltool.course.course import CourseContainer
        >>> container = CourseContainer()
        >>> importer = CourseCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Course 1, Course 1 Description
        ... Course2,,course2_local,course2_gov,5   
        ... Course3, Course 3 Description,,,10
        ... Course 4, Course 4 description, local_course4, gov_course4\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the courses exist

        >>> [course for course in container]
        [u'course-1', u'course2_local', u'course3', u'local_course4']

    Check that descriptions were imported properly

        >>> [course.title for course in container.values()]
        ['Course 1', 'Course2', 'Course3', 'Course 4']
        
        >>> [course.description for course in container.values()]
        ['Course 1 Description', '', 'Course 3 Description', 'Course 4 description']
        
        >>> [course.course_id for course in container.values()]
        [None, 'course2_local', None, 'local_course4']
        
        >>> [course.government_id for course in container.values()]
        [None, 'course2_gov', None, 'gov_course4']

        >>> [course.credits for course in container.values()]
        [None, Decimal('5'), Decimal('10'), None]

    """


def doctest_CourseCSVImporter_invalid_id():
    r"""Tests for CourseCSVImporter.

    Create a course container and an importer

        >>> from schooltool.course.browser.csvimport import CourseCSVImporter
        >>> from schooltool.course.course import CourseContainer
        >>> container = CourseContainer()
        >>> importer = CourseCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Course 1, Course 1 Description, +foo'''
        >>> importer.importFromCSV(csvdata)
        False

    We should get an error, because the id is invalid:

        >>> for error in importer.errors.fields:
        ...     print translate(error)
        Course "Course 1" id "+foo" is invalid. Names cannot begin with '+' or '@' or contain '/'

    """


def doctest_CourseCSVImporter_invalid_credits():
    r"""Tests for CourseCSVImporter.

    Create a course container and an importer

        >>> from schooltool.course.browser.csvimport import CourseCSVImporter
        >>> from schooltool.course.course import CourseContainer
        >>> container = CourseContainer()
        >>> importer = CourseCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Course 1, Course 1 Description, course1_local, , 34.3
        ... Course3,, course3, , 2+2
        ... Course2,,,,invalid'''
        >>> importer.importFromCSV(csvdata)
        False

    We should get an error, because the credits value is not a number:

        >>> for error in importer.errors.fields:
        ...     print translate(error)
        Course "Course3" credits "2+2" value must be a number.
        Course "Course2" credits "invalid" value must be a number.

    Note that 34.3 is a valid value number.

    """


def doctest_CourseCSVImporter_reimport():
    r"""Tests for CourseCSVImporter.

    Create a course container and an importer

        >>> from schooltool.course.browser.csvimport import CourseCSVImporter
        >>> from schooltool.course.course import CourseContainer
        >>> container = CourseContainer()
        >>> importer = CourseCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Course 1, Course 1 Description
        ... Course2,,course-2, course-2-gov, 10   ,
        ... Course3, Course 3 Description,,course-3-gov\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True


    Check that the courses exist

        >>> [course for course in container]
        [u'course-1', u'course-2', u'course3']

    Check that titles were imported properly

        >>> [course.title for course in container.values()]
        ['Course 1', 'Course2', 'Course3']

    Check that descriptions were imported properly

        >>> [course.description for course in container.values()]
        ['Course 1 Description', '', 'Course 3 Description']

    Check that course ids were imported properly

        >>> [course.course_id for course in container.values()]
        [None, 'course-2', None]

    Check that goverment ids were imported properly

        >>> [course.government_id for course in container.values()]
        [None, 'course-2-gov', 'course-3-gov']

    Check that credits were imported properly

        >>> [course.credits for course in container.values()]
        [None, Decimal('10'), None]

    Now import a different CSV with some courses matching:

        >>> csvdata='''Course 1, Course Description
        ... Course2,             ,course-2,,20
        ... Course4, Course 4 Description\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the courses exist

        >>> [course for course in container]
        [u'course-1', u'course-2', u'course3', u'course4']

    Check that descriptions were updated properly

        >>> [course.description for course in container.values()]
        ['Course Description', '', 'Course 3 Description', 'Course 4 Description']

    Check that credits were imported properly

        >>> [course.credits for course in container.values()]
        [None, Decimal('20'), None, None]

    By the way - ID takes precedence over title so if we import:

        >>> csvdata='''Course4, Description, course3'''
        >>> importer.importFromCSV(csvdata)
        True

    We definitely get the same amount of courses

        >>> [course for course in container]
        [u'course-1', u'course-2', u'course3', u'course4']

    But description and title of the course3 have been changed:

        >>> container['course3'].title
        'Course4'

        >>> container['course3'].description
        'Description'

    While course 4 hasn't been modified:

        >>> container['course4'].title
        'Course4'

        >>> container['course4'].description
        'Course 4 Description'

    """


def doctest_CourseCSVImportView():
    r"""
    We'll create a course csv import view

        >>> from schooltool.course.browser.csvimport import CourseCSVImportView
        >>> from schooltool.course.course import CourseContainer
        >>> from zope.publisher.browser import TestRequest
        >>> container = CourseContainer()
        >>> request = TestRequest()

    Now we'll try a text import.  Note that the description is not required

        >>> request.form = {'csvtext' : u"A Course, The best Course, some-course\nAnother Course,,,,5\nEspañol, Descripción, spanish, spanish-gov, 20\n\n",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = CourseCSVImportView(container, request)
        >>> view.update()
        >>> sorted([course for course in container])
        [u'another-course', u'some-course', u'spanish']
        >>> [container[key].course_id for key in sorted(container)]
        [None, u'some-course', u'spanish']
        >>> [container[key].government_id for key in sorted(container)]
        [None, None, u'spanish-gov']
        >>> [container[key].credits for key in sorted(container)]
        [Decimal('5'), None, Decimal('20')]

    If no data is provided, we naturally get an error

        >>> request.form = {'charset' : 'UTF-8', 'UPDATE_SUBMIT': 1}
        >>> view = CourseCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'No data provided']

    We also get an error if a line starts with a comma (no title)

        >>> request.form = {'csvtext' : u", No title provided here",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = CourseCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'Failed to import CSV text', u'Titles may not be empty']

    """


def doctest_SectionContainerView():
    r"""View for the sections container.

        >>> from schooltool.course.browser.section import SectionContainerView
        >>> from schooltool.course.section import SectionContainer
        >>> sc = SectionContainer()
        >>> request = TestRequest()
        >>> view = SectionContainerView(sc, request)

    """


def doctest_SectionView():
    r"""View for sections

    Lets create a simple view for a section:

        >>> from schooltool.course.browser.section import SectionView
        >>> from schooltool.course.section import Section
        >>> section = Section()
        >>> request = TestRequest()
        >>> view = SectionView(section, request)

    Stub the app.

        >>> from schooltool.person.person import Person, PersonContainer

        >>> class AppStub(dict):
        ...    pass
        >>> app = AppStub(persons=PersonContainer())

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(lambda x: app, (None,), ISchoolToolApplication)

    Stub a table formatter.

        >>> class TableFormatterStub(object):
        ...     def __init__(self, container, request):
        ...         self.container = container
        ...     def setUp(self, table_formatter=None, items=[], batch_size=0):
        ...         self.items = items
        ...         self.table_formatter = table_formatter
        ...     def render(self):
        ...         print 'Rendering persons from %s, using %s:' % (
        ...             self.container, self.table_formatter)
        ...         for item in self.items:
        ...             print item.title

        >>> from zope.publisher.interfaces.http import IHTTPRequest
        >>> from schooltool.person.interfaces import IPersonContainer

        >>> from schooltool.table.interfaces import ITableFormatter
        >>> provideAdapter(lambda p, r: TableFormatterStub(p, r),
        ...                adapts=(IPersonContainer, IHTTPRequest),
        ...                provides=ITableFormatter)

    Let's add some students.

        >>> person1 = Person(title='Person1')
        >>> person2 = Person(title='Person2')
        >>> person3 = Person(title='Person3')
        >>> section.members.add(person1)
        >>> section.members.add(person2)
        >>> section.members.add(person3)

    And render the person table.

        >>> view.renderPersonTable()
        Rendering persons from <...PersonContainer ...>,
        using <class 'zc.table.table.StandaloneFullFormatter'>:
        Person1
        Person2
        Person3

    """


def doctest_SectionEditView():
    r"""Test for SectionEditView

    Let's create a view for editing a section:

        >>> from schooltool.course.browser.section import SectionEditView
        >>> from schooltool.course.section import Section
        >>> from schooltool.course.interfaces import ISection

    We need some setup:

        >>> app = setup.setUpSchoolToolSite()
        >>> from schooltool.resource.resource import Location
        >>> app['resources']['room1'] = room1 = Location("Room 1")

        >>> section = Section()
        >>> directlyProvides(section, IContainmentRoot)
        >>> request = TestRequest()

        >>> class TestSectionEditView(SectionEditView):
        ...     schema = ISection
        ...     fieldNames = ('title', 'description')
        ...     _factory = Section

        >>> view = TestSectionEditView(section, request)

    We should not get redirected if we did not click on apply button:

        >>> request = TestRequest()
        >>> view = TestSectionEditView(section, request)
        >>> view.update()
        ''
        >>> request.response.getStatus()
        599

    After changing name of the section you should get redirected to the
    container:

        >>> request = TestRequest()
        >>> request.form = {'UPDATE_SUBMIT': 'Apply',
        ...                 'field.title': u'new_title'}
        >>> view = TestSectionEditView(section, request)
        >>> view.update()
        u'Updated on ${date_time}'
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

        >>> section.title
        u'new_title'

    Even if the title has not changed you should get redirected to the section
    list:

        >>> request = TestRequest()
        >>> request.form = {'UPDATE_SUBMIT': 'Apply',
        ...                 'field.title': u'new_title'}
        >>> view = TestSectionEditView(section, request)
        >>> view.update()
        ''
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

        >>> section.title
        u'new_title'

    We can cancel an action if we want to:

        >>> request = TestRequest()
        >>> request.form = {'CANCEL': 'Cancel'}
        >>> view = TestSectionEditView(section, request)
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    """


def doctest_SectionMemberCSVImporter():
    r"""Tests for SectionMemberCSVImporter.

    First we need to set up some persons:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> from schooltool.course.section import PersonLearnerAdapter
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> smith = persons['smith'] = Person('smith', 'John Smith')
        >>> [section.title for section in PersonLearnerAdapter(smith).sections()]
        []
        >>> jones = persons['jones'] = Person('jones', 'Sally Jones')
        >>> [section.title for section in PersonLearnerAdapter(jones).sections()]
        []
        >>> stevens = persons['stevens'] = Person('stevens', 'Bob Stevens')
        >>> [section.title for section in PersonLearnerAdapter(stevens).sections()]
        []

    Create a section and an importer

        >>> from schooltool.course.browser.csvimport import SectionMemberCSVImporter
        >>> from schooltool.course.section import Section
        >>> section = Section('Section title', 'Section description')
        >>> [person.username for person in section.members]
        []
        >>> importer = SectionMemberCSVImporter(section, None)

    Import some sample data

        >>> csvdata='''smith
        ... stevens\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the persons were added to the section members:

        >>> [person.username for person in section.members]
        ['smith', 'stevens']
        >>> [section.title for section in PersonLearnerAdapter(smith).sections()]
        ['Section title']
        >>> [section.title for section in PersonLearnerAdapter(jones).sections()]
        []
        >>> [section.title for section in PersonLearnerAdapter(stevens).sections()]
        ['Section title']

    Create another section and another importer

        >>> another_section = Section('Another section', 'Another description')
        >>> [person.username for person in another_section.members]
        []
        >>> another_importer = SectionMemberCSVImporter(another_section, None)

    Import some more data

        >>> csvdata='''stevens
        ... jones\n\n\n'''
        >>> another_importer.importFromCSV(csvdata)
        True

    Check that the persons were added to the another section members:

        >>> [person.username for person in another_section.members]
        ['stevens', 'jones']
        >>> [section.title for section in PersonLearnerAdapter(smith).sections()]
        ['Section title']
        >>> [section.title for section in PersonLearnerAdapter(jones).sections()]
        ['Another section']
        >>> [section.title for section in PersonLearnerAdapter(stevens).sections()]
        ['Section title', 'Another section']

    """


def doctest_SectionMemberCSVImportView():
    r"""Tests for SectionMemberCSVImportView

    First we need to set up some persons:

        >>> from zope.i18n import translate
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> smith = persons['smith'] = Person('smith', 'John Smith')
        >>> jones = persons['jones'] = Person('jones', 'Sally Jones')
        >>> stevens = persons['stevens'] = Person('stevens', 'Bob Stevens')

    We'll create a section member csv import view

        >>> from schooltool.course.browser.csvimport import \
        ...      SectionMemberCSVImportView
        >>> from schooltool.course.section import Section
        >>> from zope.publisher.browser import TestRequest
        >>> section = Section('Section title', 'Section description')
        >>> request = TestRequest()

    Now we'll try a text import.

        >>> request.form = {
        ...     'csvtext' : u'stevens\n',
        ...     'charset' : 'UTF-8',
        ...     'UPDATE_SUBMIT': 1}
        >>> view = SectionMemberCSVImportView(section, request)
        >>> view.update()
        >>> [person.username for person in section.members]
        ['stevens']

    If no data is provided, we naturally get an error

        >>> request.form = {'charset' : 'UTF-8', 'UPDATE_SUBMIT': 1}
        >>> view = SectionMemberCSVImportView(section, request)
        >>> view.update()
        >>> view.errors
        [u'No data provided']

    We also get an error if a line doesn't have a username

        >>> request.form = {'csvtext' : u" ,stevens\njones,Sally",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = SectionMemberCSVImportView(section, request)
        >>> view.update()
        >>> view.errors
        [u'Failed to import CSV text', u'User names must not be empty.']

    Or if the username is not in the persons container

        >>> request.form = {'csvtext' : u"foobar\nstevens\njones",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = SectionMemberCSVImportView(section, request)
        >>> view.update()
        >>> [translate(error) for error in view.errors]
        [u'Failed to import CSV text', u'"foobar" is not a valid username.']

    """


def doctest_SectionInstructorView():
    """Tests for SectionInstructorView.

    First we need to set up some persons:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> persons['smith'] = Person('smith', 'John Smith')
        >>> persons['jones'] = Person('jones', 'Sally Jones')
        >>> persons['stevens'] = Person('stevens', 'Bob Stevens')

    getCollection plainly returns instructors attribute of a section:

        >>> from schooltool.course.browser.section import SectionInstructorView
        >>> class SectionStub(object):
        ...     instructors = [persons['smith']]
        >>> view = SectionInstructorView(SectionStub(), None)
        >>> [item.title for item in view.getCollection()]
        ['John Smith']

    All persons that are not in the current instructor list are
    considered available:

        >>> [item.title for item in view.getAvailableItems()]
        ['Sally Jones', 'Bob Stevens']

        >>> view.context.instructors = []

        >>> [item.title for item in view.getAvailableItems()]
        ['Sally Jones', 'John Smith', 'Bob Stevens']

    """


def doctest_SectionLearnerView():
    """Tests for SectionLearnerView.

    First we need to set up some persons:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> smith = persons['smith'] = Person('smith', 'John Smith')
        >>> jones = persons['jones'] = Person('jones', 'Sally Jones')
        >>> stevens = persons['stevens'] = Person('stevens', 'Bob Stevens')

    getCollection plainly returns members attribute of a section:

        >>> from schooltool.course.browser.section import SectionLearnerView
        >>> class SectionStub(object):
        ...     members = [smith]
        >>> view = SectionLearnerView(SectionStub(), None)
        >>> [item.title for item in view.getCollection()]
        ['John Smith']

    All persons that are not in the selected learner list are
    considered available:

        >>> [item.title for item in view.getAvailableItems()]
        ['Sally Jones', 'Bob Stevens']

        >>> view.context.members = []

        >>> [item.title for item in view.getAvailableItems()]
        ['Sally Jones', 'John Smith', 'Bob Stevens']

    Any non-person members should be skipped when displaying selected
    items:

        >>> from schooltool.group.group import Group
        >>> frogs = Group('frogs')
        >>> view.context.members = [smith, frogs]
        >>> [item.title for item in view.getSelectedItems()]
        ['John Smith']

    """


def doctest_CoursesViewlet():
    r"""Test for CoursesViewlet

    Adapters:

        >>> from zope.ucol.localeadapter import LocaleCollator
        >>> from zope.component import provideAdapter, provideUtility
        >>> from zope.i18n.interfaces.locales import ICollator
        >>> from zope.i18n.interfaces.locales import ILocale
        >>> from schooltool.course.section import PersonInstructorAdapter
        >>> from schooltool.course.section import PersonLearnerAdapter
        >>> from schooltool.course.interfaces import ISection
        >>> from schooltool.term.interfaces import ITerm
        >>> from schooltool.schoolyear.interfaces import ISchoolYear
        >>> provideAdapter(LocaleCollator, [ILocale], ICollator)
        >>> provideAdapter(PersonInstructorAdapter)
        >>> provideAdapter(PersonLearnerAdapter)
        >>> class DummySchoolYear(object):
        ...     title = '2010'
        ...     first = 'date'
        >>> class DummyTerm(object):
        ...     title = 'Semester'
        ...     first = 'date'
        >>> year2010 = DummySchoolYear()
        >>> semester = DummyTerm()
        >>> def DummySchoolYearSectionAdapter(section):
        ...     return year2010
        >>> def DummyTermSectionAdapter(section):
        ...     return semester
        >>> provideAdapter(DummySchoolYearSectionAdapter, [ISection], ISchoolYear)
        >>> provideAdapter(DummyTermSectionAdapter, [ISection], ITerm)

    Let's create a viewlet for a person's courses:

        >>> from schooltool.course.browser.course import CoursesViewlet
        >>> from schooltool.person.person import Person

        >>> school = setup.setUpSchoolToolSite()
        >>> persons = school['persons']
        >>> from schooltool.course.section import SectionContainer
        >>> sections = SectionContainer()

        >>> persons['teacher'] = teacher = Person("Teacher")
        >>> teacher_view = CoursesViewlet(teacher, TestRequest(), None, None)

    Not a teacher yet:

        >>> teacher_view.update()
        >>> teacher_view.isTeacher()
        False

    We'll need something to teach:

        >>> from schooltool.course.section import Section
        >>> sections['section'] = section1 = Section(title="History")
        >>> sections['section2'] = section2 = Section(title="Algebra")
        >>> section1.instructors.add(teacher)

    Now we're teaching:

        >>> teacher_view.update()
        >>> teacher_view.isTeacher()
        True
        >>> teacher_view.isLearner()
        False

    We'll teach 2 courses this semester, and we'll need an easy way to get a
    list of all the courses we're teaching.

        >>> section2.instructors.add(teacher)
        >>> teacher_view = CoursesViewlet(teacher, TestRequest(), None, None)
        >>> teacher_view.update()
        >>> for schoolyear_data in teacher_view.instructorOf:
        ...     for term in schoolyear_data['terms']:
        ...         for section in term['sections']:
        ...             print section['obj'].title
        Algebra
        History

    Let's create a student

        >>> persons['student'] = student = Person("Student")
        >>> student_view = CoursesViewlet(student, TestRequest(), None, None)
        >>> student_view.update()

        >>> student_view.isTeacher()
        False
        >>> student_view.isLearner()
        False

    Membership in a Section implies being a learner:

        >>> section2.members.add(student)
        >>> student_view.update()
        >>> student_view.isTeacher()
        False

        >>> student_view.isLearner()
        True

        >>> sections['section3'] = section3 = Section(title="English")
        >>> sections['section4'] = section4 = Section(title="Gym")

    Our student is taking several classes

        >>> section3.members.add(student)
        >>> student_view = CoursesViewlet(student, TestRequest(), None, None)
        >>> student_view.update()

        >>> for schoolyear_data in student_view.learnerOf:
        ...     for term in schoolyear_data['terms']:
        ...         for section in term['sections']:
        ...             print section['obj'].title
        Algebra
        English

    """


def test_suite():
    suite = unittest.TestSuite()
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetabling calendar integration.
"""
from datetime import datetime, time, timedelta
import urllib

import zope.schema
from zope.app.form.browser.add import AddView
from zope.interface import Interface, implements, implementer
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import CachedProperty
from zope.component import adapter, getUtility
from zope.formlib import form
from zope.html.field import HtmlFragment
from zope.session.interfaces import ISession
from zope.proxy import sameProxiedObjects
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.viewlet.interfaces import IViewlet

from schooltool.app.browser import ViewPreferences
from schooltool.app.browser.interfaces import IEventForDisplay
from schooltool.app.browser.cal import CalendarEventView
from schooltool.app.browser.cal import CalendarEventViewMixin
from schooltool.app.browser.cal import YearlyCalendarView
from schooltool.app.browser.cal import DailyCalendarRowsView
from schooltool.app.browser.cal import getCalendarEventDeleteLink
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.membership import URIMembership, URIGroup
from schooltool.app.utils import vocabulary
from schooltool.app.relationships import URISection, URIInstruction
from schooltool.calendar.browser.event import FlourishCalendarEventAddView
from schooltool.person.interfaces import IPerson
from schooltool.relationship import getRelatedObjects
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool.timetable import interfaces
from schooltool.timetable.calendar import ScheduleCalendarEvent
from schooltool.timetable.schedule import iterMeetingsInTimezone
from schooltool.timetable.interfaces import IHaveSchedule
from schooltool.term.interfaces import IDateManager, ITermContainer
from schooltool.term.term import getTermForDate

from schooltool.common import SchoolToolMessage as _


scheduleCalendarFieldNames = (
    "title", "description", "start_date", "start_time",
    "duration", "duration_type", "allday", "location",
    )


class ScheduleEventEditView(CalendarEventView, form.Form):

    title = _("Modify meeting information")

    form_fields = form.fields(HtmlFragment(
            __name__='description',
            title=_("Description"),
            required=False))

    template = ViewPageTemplateFile("templates/schedule_event_edit.pt")

    def setUpWidgets(self, ignore_request=False):
        self.widgets = form.setUpEditWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            ignore_request=ignore_request)

    def __init__(self, context, request):
        form.Form.__init__(self, context, request)
        CalendarEventView.__init__(self, context, request)

    def redirect_to_parent(self):
        url = absoluteURL(self.context.__parent__, self.request)
        self.request.response.redirect(url)
        return ''

    @form.action(_("Apply"))
    def handle_edit_action(self, action, data):
        self.context.description = data['description']
        return self.redirect_to_parent()

    @form.action(_("Cancel"), condition=form.haveInputWidgets)
    def handle_cancel_action(self, action, data):
        return self.redirect_to_parent()


class FlourishScheduleEventEditView(flourish.page.Page,
                                    ScheduleEventEditView):
    __init__ = ScheduleEventEditView.__init__
    update = ScheduleEventEditView.update

    legend = _("Meeting Details")

    def nextURL(self):
        return (self.request.get('back_url') or
                absoluteURL(self.context, self.request))


class IScheduleEventAddForm(Interface):
    """Schema for schedule calendar event adding form."""

    title = zope.schema.TextLine(
        title=_("Title"),
        required=False)
    allday = zope.schema.Bool(
        title=_("All day"),
        required=False)
    start_date = zope.schema.Date(
        title=_("Date"),
        required=False)
    start_time = zope.schema.TextLine(
        title=_("Time"),
        description=_("Start time in 24h format"),
        required=False)

    duration = zope.schema.Int(
        title=_("Duration"),
        required=False,
        default=60)

    duration_type = zope.schema.Choice(
        title=_("Duration Type"),
        required=False,
        default="minutes",
        vocabulary=vocabulary([("minutes", _("Minutes")),
                               ("hours", _("Hours")),
                               ("days", _("Days"))]))

    location = zope.schema.TextLine(
        title=_("Location"),
        required=False)

    description = HtmlFragment(
        title=_("Description"),
        required=False)


class ScheduleEventAddView(CalendarEventViewMixin, AddView):
    """A view for adding an event."""

    schema = IScheduleEventAddForm

    title = _("Add meeting")
    submit_button_title = _("Add")

    show_book_checkbox = True
    show_book_link = False
    _event_uid = None

    error = None

    def __init__(self, context, request):

        prefs = ViewPreferences(request)
        self.timezone = prefs.timezone

        if "field.start_date" not in request:
            # XXX: shouldn't use date.today; it depends on the server's timezone
            # which may not match user expectations
            today = getUtility(IDateManager).today.strftime("%Y-%m-%d")
            request.form["field.start_date"] = today
        super(AddView, self).__init__(context, request)

    def create(self, **kwargs):
        """Create an event."""
        data = self.processRequest(kwargs)
        event = self._factory(data['start'], data['duration'], data['title'],
                              location=data['location'],
                              allday=data['allday'],
                              description=data['description'])
        # XXX: also meeting id! Don't forget the meeting id.
        return event

    def add(self, event):
        self.context.addEvent(event)
        uid = event.unique_id
        self._event_name = event.__name__
        session_data = ISession(self.request)['schooltool.calendar']
        session_data.setdefault('added_event_uids', set()).add(uid)
        return event

    def update(self):
        if 'UPDATE' in self.request:
            return self.updateForm()
        elif 'CANCEL' in self.request:
            self.update_status = ''
            self.request.response.redirect(self.nextURL())
            return self.update_status
        else:
            return AddView.update(self)

    def nextURL(self):
        if "field.book" in self.request:
            url = absoluteURL(self.context, self.request)
            return '%s/%s/booking.html' % (url, self._event_name)
        else:
            return absoluteURL(self.context, self.request)


class FlourishScheduleEventAddView(FlourishCalendarEventAddView):

    fieldNames = scheduleCalendarFieldNames
    schema = IScheduleEventAddForm

    _keyword_arguments = scheduleCalendarFieldNames
    _factory = ScheduleCalendarEvent

    def create(self, **kwargs):
        data = self.processRequest(kwargs)
        event = self._factory(data['start'], data['duration'], data['title'],
                              location=data['location'],
                              allday=data['allday'],
                              description=data['description'])
        # XXX: also meeting id! Don't forget the meeting id.
        return event

    def setUpCustomWidgets(self):
        self.setCustomWidget('description', height=5)

    def update(self):
        if ("field.title" not in self.request):
            calendar = ISchoolToolCalendar(self.context)
            owner = IHaveSchedule(calendar.__parent__)
            self.request.form["field.title"] = owner.title
        super(FlourishScheduleEventAddView, self).update()


class TermLegendViewlet(object):
    implements(IViewlet)

    def legend(self):
        terms = self.__parent__.legend.items()
        terms.sort(key=lambda t: t[0].first)
        return [{'title': term.title,
                 'cssclass': "legend-item term%s" % cssClass}
                for term, cssClass in terms]


class ScheduleYearlyCalendarView(YearlyCalendarView):

    def __init__(self, context, request):
        super(YearlyCalendarView, self).__init__(context,request)
        self.numterms = 1
        self.calendar = None

    @CachedProperty
    def legend(self):
        numterms = 1
        legend = {}
        terms = ITermContainer(self.cursor, {})
        for quarter in self.getYear(self.cursor):
            for month in quarter:
                for week in month:
                    for day in week:
                        term = None
                        for term in terms.values():
                            if day.date in term:
                                break
                        if term and not term in legend:
                            legend[term] = self.numterms
                            numterms += 1
        return legend

    def renderRow(self, week, month):
        result = []

        terms = ITermContainer(self.cursor, {})

        for day in week:
            term = None
            for term in terms.values():
                if day.date in term:
                    break
            cssClass = "term%d" % self.legend.get(term, 0)
            result.append('<td class="cal_yearly_day">')
            if day.date.month == month:
                if day.today():
                    cssClass += ' today'
                # Let us hope that URLs will not contain < > & or "
                # This is somewhat related to
                # https://bugs.launchpad.net/schooltool/+bug/79781
                result.append('<a href="%s" class="%s">%s</a>' %
                              (self.calURL('daily', day.date), cssClass,
                               day.date.day))
            result.append('</td>')

        return "\n".join(result)


class ScheduleDailyCalendarRowsView(DailyCalendarRowsView):
    """Daily calendar rows view for SchoolTool.

    This view differs from the original view in SchoolTool in that it can
    also show schedule periods instead of hour numbers.
    """

    def getDefaultMeetings(self, date, timezone):
        term = getTermForDate(date)
        if term is None:
            return []
        schoolyear = ISchoolYear(term)
        timetable_container = interfaces.ITimetableContainer(schoolyear)
        default_schedule = timetable_container.default
        if default_schedule is None:
            return []
        meetings = list(iterMeetingsInTimezone(
                default_schedule, timezone, date))
        return meetings

    def meetingTitle(self, meeting):
        if (meeting.period is None or
            not meeting.period.title):
            return self.rowTitle(meeting.dtstart.time(), meeting.duration)
        return meeting.period.title

    def calendarRows(self, cursor, starthour, endhour, events):
        tz = self.getPersonTimezone()
        meetings = self.getDefaultMeetings(cursor, tz.zone)
        meeting_rows = [
            (self.meetingTitle(meeting), meeting.dtstart, meeting.duration)
            for meeting in meetings]

        daystart = tz.localize(datetime.combine(cursor, time()))

        rows = []

        row_start = tz.localize(datetime.combine(cursor, time()) +
                                timedelta(hours=starthour))
        row_start = tz.normalize(row_start)

        rows_end = tz.localize(datetime.combine(cursor, time()) +
                               timedelta(hours=endhour))
        rows_end = tz.normalize(rows_end + timedelta(hours=1))

        if meeting_rows:
            row_start = min(row_start, meeting_rows[0][1].astimezone(tz))
            meet_end = meeting_rows[-1][1] + meeting_rows[-1][2]
            rows_end = max(rows_end, meet_end.astimezone(tz))

        while row_start < rows_end:
            row = None
            if meeting_rows:
                time_until_meeting = meeting_rows[0][1] - row_start
                if time_until_meeting < time.resolution:
                    meeting = meeting_rows.pop(0)
                    row = meeting
                elif time_until_meeting < timedelta(hours=1, minutes=15):
                    row = (self.rowTitle(row_start.time(), time_until_meeting),
                           row_start, time_until_meeting)

            if row is None:
                hstart = tz.normalize(row_start + timedelta(hours=1))
                hstart = hstart.replace(minute=0, second=0, microsecond=0)
                delta = hstart - row_start
                if delta < timedelta(minutes=15):
                    delta += timedelta(hours=1)
                row = (self.rowTitle(row_start.time(), delta), row_start, delta)

            if row[1].astimezone(tz).date() == daystart.date():
                rows.append(row)
            row_start = tz.normalize(row_start + row[2])
        return rows


@adapter(IEventForDisplay, IBrowserRequest, interfaces.IScheduleCalendar)
@implementer(Interface)
def getScheduleCalendarEventDeleteLink(event, request, calendar):
    schedule = event.context.schedule
    if schedule is not None:
        return None
    return getCalendarEventDeleteLink(event, request, calendar)


class TimetableCalendarListSubscriber(object):
    """A subscriber that can tell which calendars should be displayed."""

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def getCalendars(self):
        """Get a list of calendars to display.

        Yields tuples (calendar, color1, color2).
        """

        owner = self.context.__parent__

        user = IPerson(self.request.principal, None)
        if (user is not None and
            sameProxiedObjects(user, owner)):
            return

        instructs = list(getRelatedObjects(
                owner, URISection, rel_type=URIInstruction))
        member_of = list(getRelatedObjects(
                owner, URIGroup, rel_type=URIMembership))

        for obj in instructs + member_of:
            if IHaveSchedule.providedBy(obj):
                cal = ISchoolToolCalendar(obj, None)
                if cal is not None:
                    yield(ISchoolToolCalendar(obj), '#9db8d2', '#7590ae')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 41.

After permissions remap, add superuser to managers and clerks in all years.
"""

from zope.app.generations.utility import getRootFolder, findObjectsProviding
from zope.component.hooks import getSite, setSite

from schooltool.app.membership import Membership
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.group.interfaces import IGroupContainer


def makeManager(app, schoolyear, person):
    groups = IGroupContainer(schoolyear)

    default_admin_names = ('manager', 'clerks')
    admin_groups = [groups.get(name)
                    for name in default_admin_names
                    if name in groups]

    person_groups = [group for group in Membership.query(member=person)
                     if group.__parent__.__name__ == groups.__name__]

    for group in admin_groups:
        if group not in person_groups:
            Membership(group=group, member=person)


def evolve(context):
    root = getRootFolder(context)

    old_site = getSite()
    apps = list(findObjectsProviding(root, ISchoolToolApplication))
    for app in apps:
        setSite(app)
        persons = ISchoolToolApplication(None)['persons']
        manager = persons.super_user
        if manager is None:
            continue
        syc = ISchoolYearContainer(app)
        for sy in syc.values():
            makeManager(app, sy, manager)

    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool skin containers
"""
import urllib

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import queryMultiAdapter
from zope.security.checker import canAccess
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.skin.flourish.page import Page
from schooltool.common import SchoolToolMessage as _


class ContainerDeleteView(Page):
    """A view for deleting items from container."""

    @property
    def container(self):
        return self.context

    def listIdsForDeletion(self):
        return [key for key in self.container
                if "delete.%s" % key in self.request]

    def _listItemsForDeletion(self):
        return [self.container[key] for key in self.listIdsForDeletion()]

    itemsToDelete = property(_listItemsForDeletion)

    def update(self):
        if 'CONFIRM' in self.request:
            for key in self.listIdsForDeletion():
                del self.container[key]
            self.request.response.redirect(self.nextURL())
        elif 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.container, self.request)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for schooltool.basicperson.sampledata
"""

import unittest
import doctest

from zope.app.testing import setup
from zope.component import provideAdapter
from zope.interface import Interface
from zope.interface.verify import verifyObject

from schooltool.group.interfaces import IGroupContainer
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.testing import setup as stsetup


def doctest_SampleStudents():
    """A sample data plugin that generates students

        >>> from schooltool.basicperson.sampledata import SampleStudents
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> plugin = SampleStudents()
        >>> verifyObject(ISampleDataPlugin, plugin)
        True

    This plugin simply creates a 1000 persons with random names and
    usernames of a form 'student123'.

        >>> plugin.power
        1000

        >>> app = stsetup.setUpSchoolToolSite()

        >>> students = IGroupContainer(app)['students']

        >>> len(students.members)
        0

        >>> plugin.generate(app, 42)

        >>> len(students.members)
        1000

        >>> for i in range(5):
        ...     print app['persons']['student%03d' % i].title
        Freja Freeman
        Klara Phillips
        Cath Rodriguez
        Daniela Petersen
        Gabbie Cunningham

    The students get their passwords set the same as their logins:

        >>> for i in range(5):
        ...     login = 'student%03d' % i
        ...     assert app['persons'][login].checkPassword(login)

    """


def doctest_SampleTeachers():
    """A sample data plugin that generates teachers

        >>> from schooltool.basicperson.sampledata import SampleTeachers
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> plugin = SampleTeachers()
        >>> verifyObject(ISampleDataPlugin, plugin)
        True

    This plugin creates this many teachers:

        >>> plugin.power
        48

    This plugin creates a number of teachers and adds them to the
    Teachers group.

        >>> app = stsetup.setUpSchoolToolSite()

        >>> teachers = IGroupContainer(app)['teachers']

        >>> len(teachers.members)
        0

        >>> plugin.generate(app, 42)
        >>> len(teachers.members)
        48

    Teachers' names are different from the students' names above:

        >>> for i in range(5):
        ...     teacher = app['persons']['teacher%03d' % i]
        ...     print teacher.title
        Catherine Martin
        Kimmy Cooper
        Helen Patterson
        Iris Laurent
        Philippa Stewart

    The teachers get their passwords set the same as their logins:

        >>> for i in range(5):
        ...     login = 'teacher%03d' % i
        ...     assert app['persons'][login].checkPassword(login)

    """


def doctest_SamplePersonalEvents():
    """A sample data plugin that generates random personal events.

        >>> from schooltool.basicperson.sampledata import SamplePersonalEvents
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> plugin = SamplePersonalEvents()
        >>> verifyObject(ISampleDataPlugin, plugin)
        True

        >>> app = stsetup.setUpSchoolToolSite()

        >>> from schooltool.basicperson.sampledata import SampleStudents
        >>> from schooltool.basicperson.sampledata import SampleTeachers
        >>> from schooltool.term.sampledata import SampleTerms
        >>> plugin_students = SampleStudents()
        >>> plugin_students.power = 20
        >>> plugin_teachers = SampleTeachers()
        >>> plugin_teachers.power = 3
        >>> plugin_terms = SampleTerms()
        >>> plugin_terms.generate(app, 42)
        >>> plugin_students.generate(app, 42)
        >>> plugin_teachers.generate(app, 42)

    Probability of person having event on any day in percents:

        >>> plugin.probability
        2

        >>> plugin.probability = 50

    Create random events for all students and teachers.

        >>> plugin.generate(app, 42)

        >>> for i in range(5):
        ...     person = app['persons']['student%03d' % i]
        ...     calendar = ISchoolToolCalendar(person)
        ...     print len(calendar)
        233
        250
        227
        252
        248

        >>> person = app['persons']['teacher000']
        >>> calendar = ISchoolToolCalendar(person)
        >>> len(calendar)
        240

        >>> events = list(calendar)
        >>> events.sort()
        >>> for event in events[0:5]:
        ...     print event.dtstart,
        ...     print event.duration,
        ...     print event.title
        2005-08-24 12:00:00+00:00 5:00:00 Tribal dances
        2005-08-25 17:30:00+00:00 1:30:00 Dentist
        2005-08-27 11:30:00+00:00 4:00:00 Concert
        2005-08-28 23:30:00+00:00 3:30:00 Quake tournament
        2005-08-29 19:00:00+00:00 6:00:00 Boating
    """


def setUp(test):
    setup.placefulSetUp()
    from schooltool.term.term import getTermContainer
    from schooltool.term.interfaces import ITermContainer
    from schooltool.schoolyear.schoolyear import getSchoolYearContainer
    provideAdapter(getTermContainer, [Interface], ITermContainer)
    provideAdapter(getSchoolYearContainer)

    from schooltool.group.group import GroupContainer, Group
    groups = GroupContainer()
    provideAdapter(lambda context: groups,
                   adapts=[ISchoolToolApplication],
                   provides=IGroupContainer)
    groups['teachers'] = Group('Teachers')
    groups['students'] = Group('Students')

    from zope.annotation.interfaces import IAnnotatable
    from schooltool.relationship.interfaces import IRelationshipLinks
    from schooltool.relationship.annotatable import getRelationshipLinks
    provideAdapter(getRelationshipLinks, [IAnnotatable], IRelationshipLinks)

    from schooltool.app.cal import getCalendar
    from schooltool.app.interfaces import ISchoolToolCalendar
    from schooltool.person.interfaces import IPerson
    provideAdapter(getCalendar, [IPerson], ISchoolToolCalendar)


def tearDown(test):
    setup.placefulTearDown()


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown)
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import datetime

import celery.states
import celery.task

from schooltool.common import SchoolToolMessage as _


IN_PROGRESS = 'IN_PROGRESS'
COMMITTING = 'COMMITTING_ZODB'


def normalized_progress(*args):
    pmin = 0.0
    pmax = 1.0
    while args:
        val, max = args.pop()
        pmin = pmin + pmax * val
        pmax = pmax * max
    return min(float(pmin) / float(pmax), 1.0)


class TaskReadState(object):

    task_id = None
    _meta = None
    _progress_states = (celery.states.STARTED, IN_PROGRESS)

    def __init__(self, task_id):
        self.task_id = task_id
        self._meta = None, None, None
        self.reload()

    def reload(self):
        result = celery.task.Task.AsyncResult(self.task_id)
        self._meta = result.state, result.result, result.traceback

    @property
    def state(self):
        return self._meta[0]

    @property
    def in_progress(self):
        return (self.state in self._progress_states or self.committing)

    @property
    def committing(self):
        return self.state == COMMITTING

    @property
    def pending(self):
        return self.state in celery.states.UNREADY_STATES

    @property
    def finished(self):
        return self.state in celery.states.READY_STATES

    @property
    def failed(self):
        return self.state in celery.states.PROPAGATE_STATES

    @property
    def succeeded(self):
        return self.state == celery.states.SUCCESS

    @property
    def progress(self):
        if self.in_progress:
            return self._meta[1]

    @property
    def result(self):
        if self.succeeded:
            return self._meta[1]

    @property
    def failure(self):
        if self.failed:
            return self._meta[1]

    @property
    def info(self):
        return self._meta[1]

    @property
    def traceback(self):
        return self._meta[2]


class NotInProgress(Exception):
    pass


class TaskWriteState(TaskReadState):

    def set_progress(self, progress=None):
        result = celery.task.Task.AsyncResult(self.task_id)
        # XXX: only check this if task.track_started
        if result.state not in self._progress_states:
            raise NotInProgress(result.state, self._progress_states)
        result.backend.store_result(result.task_id, progress, IN_PROGRESS)
        self.reload()

    def set_committing(self):
        result = celery.task.Task.AsyncResult(self.task_id)
        # XXX: only check this if task.track_started
        if result.state not in self._progress_states:
            raise NotInProgress(result.state, self._progress_states)
        result.backend.store_result(result.task_id, self.info, COMMITTING)
        self.reload()


class Timer(object):

    update_interval = datetime.timedelta(seconds=1)
    last_updated = None

    def __init__(self):
        self.reset()

    def reset(self):
        self.last_updated = None

    @property
    def now(self):
        return datetime.datetime.utcnow()

    @property
    def delta(self):
        last_updated = self.last_updated
        if last_updated is None:
            return None
        return self.now - last_updated

    def tick(self, *args, **kw):
        raise NotImplemented()

    def tock(self, *args, **kw):
        raise NotImplemented()

    def force(self, *args, **kw):
        self.tick(*args, **kw)
        self.tock(*args, **kw)
        self.last_updated = self.now

    def __call__(self, *args, **kw):
        self.tick(*args, **kw)
        if (self.last_updated is None or
            self.delta >= self.update_interval):
            self.tock(*args, **kw)
            self.last_updated = self.now


class Progress(Timer):

    title = u''
    value = 0.0
    max_value = 1.0

    def __init__(self, default_title=u'', max=None):
        self.default_title = default_title
        self.max_value = max or 1.0

    def reset(self):
        self.title = self.default_title
        self.value = 0.0

    def tick(self, title=None, progress=None):
        if title is not None:
            self.title = title
        if progress is not None:
            self.progress = progress

    def tock(self, *args, **kw):
        pass


class TaskProgress(Timer):

    parts = None
    order = None
    value = None
    task_status = None

    def __init__(self, task_id):
        self.parts = {}
        self.order = []
        self.task_status = TaskWriteState(task_id)
        Timer.__init__(self)

    def reset(self):
        Timer.reset(self)
        self.value = {}
        for n, importer in enumerate(self.parts):
            self.value[n] = {
                'title': importer.title,
                'errors': [],
                'progress': 0.0,
                }
        #self.value['overall'] = {
        #        'title': _('Overall'),
        #        'errors': [],
        #        'progress': 0.0,
        #        }
        self.tock()

    def finish(self):
        for status in self.value.values():
            status['progress'] = 1.0
        self.task_status.set_progress(self.value)
        self.last_updated = self.now

    def tick(self, part_name, title=None, progress=None):
        if part_name not in self.parts:
            self.parts[part_name] = Progress(default_title=title)
        
            
        self.value[importer_n]['progress'] = value
        self.value['overall']['progress'] = normalized_progress(
            importer_n, len(self.importers), value, 1.0)

    def tock(self, *args, **kw):
        self.task_status.set_progress(self.value)





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for Person views.
"""
import unittest
import doctest

from zope.publisher.browser import TestRequest

from schooltool.app.browser.testing import setUp, tearDown


def doctest_PersonPhotoView():
    r"""Test for PersonPhotoView

    We will need a person that has a photo:

        >>> from schooltool.person.person import Person
        >>> person = Person()
        >>> person.photo = "I am a photo!"

    We can now create a view:

        >>> from schooltool.person.browser.person import PersonPhotoView
        >>> request = TestRequest()
        >>> view = PersonPhotoView(person, request)

    The view returns the photo and sets the appropriate Content-Type header:

        >>> view()
        'I am a photo!'
        >>> request.response.getHeader("Content-Type")
        'image/jpeg'

    However, if a person has no photo, the view raises a NotFound error.

        >>> person.photo = None
        >>> view()                                  # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        NotFound: Object: <...Person object at ...>, name: u'photo'

    """


def doctest_PersonFilterWidget():
    """Doctest for PersonFilterWidget.

    For this test we will need a catalog with an index for person
    titles:

        >>> from zope.catalog.interfaces import ICatalog
        >>> class IndexStub(object):
        ...     def __init__(self):
        ...         self.documents_to_values = {}
        ...     def apply(self, query):
        ...         query = query.replace('*', '')
        ...         results = []
        ...         for id, value in self.documents_to_values.items():
        ...             if query in value:
        ...                 results.append(id)
        ...         return results
        >>> text_index = IndexStub()

        >>> class CatalogStub(dict):
        ...     def __init__(self):
        ...         self['text'] = text_index
        >>> catalog = CatalogStub()

   Some persons:

        >>> class PersonStub(object):
        ...     def __init__(self, title, groups, person_id):
        ...         self.id = person_id
        ...         self.title = title
        ...         self.__name__ = title
        ...         for group in groups:
        ...             group.add(self)
        ...     def __repr__(self):
        ...         return '<ItemStub %s>' % self.title

   Some groups:

        >>> class GroupStub(object):
        ...     def __init__(self, title):
        ...         self.title = title
        ...         self.members = []
        ...     def add(self, member):
        ...         self.members.append(member)
        >>> a = GroupStub('a')
        >>> b = GroupStub('b')
        >>> c = GroupStub('c')

   Container with some persons in it:

        >>> class ContainerStub(dict):
        ...     def __init__(self):
        ...         persons = [('a1234','alpha', [a]),
        ...                    ('a1235','beta', [b, c]),
        ...                    ('a1236','lambda', [b])]
        ...         for id, (username, title, groups) in enumerate(persons):
        ...             self[username] = PersonStub(title, groups, id)
        ...             text_index.documents_to_values[id] = ' '.join([title, username])
        ...     def __conform__(self, iface):
        ...         if iface == ICatalog:
        ...             return catalog

    Let's create the PersonFilterWidget:

        >>> from zope.publisher.browser import TestRequest
        >>> from schooltool.person.browser.person import PersonFilterWidget
        >>> container = ContainerStub()
        >>> request = TestRequest()
        >>> widget = PersonFilterWidget(container, request)

    The state of the widget (whether it will filter the data or not)
    is determined by checking whether there is at least one query
    parameter in the request:

        >>> widget.active()
        False

        >>> request.form = {'SEARCH_TITLE': 'lamb'}
        >>> widget.active()
        True

        >>> request.form = {'SEARCH_GROUP': 'lamb'}
        >>> widget.active()
        True

    The information that we got from the request can be appended to
    the url:

        >>> widget.extra_url()
        '&SEARCH_GROUP=lamb'

        >>> request.form = {'SEARCH_TITLE': 'lamb', 'SEARCH_GROUP': 'a'}
        >>> widget.extra_url()
        '&SEARCH_TITLE=lamb&SEARCH_GROUP=a'

    Filtering is done by skipping any entry that doesn't contain the
    query string in it's title, or are not in the target group:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from zope.component import adapts
        >>> from zope.interface import implements, Interface
        >>> class StubApplication(dict):
        ...     implements(ISchoolToolApplication)
        ...     adapts(Interface)
        ...     def __init__(self, context):
        ...         self['groups'] = {'a': a, 'b': b, 'c': c}
        >>> from zope.component import provideAdapter
        >>> provideAdapter(StubApplication)
        >>> from schooltool.group.interfaces import IGroupContainer
        >>> provideAdapter(lambda app: app['groups'],
        ...                adapts=[ISchoolToolApplication],
        ...                provides=IGroupContainer)

        >>> items = [{'id': 0},
        ...          {'id': 1},
        ...          {'id': 2}]

        >>> request.form = {'SEARCH_TITLE': 'lamb'}
        >>> widget.filter(items)
        [{'id': 2}]

        >>> from zope.component import provideUtility
        >>> from zope.intid.interfaces import IIntIds
        >>> class IntIdsStub(object):
        ...     def queryId(self, obj):
        ...         return obj.id
        >>> provideUtility(IntIdsStub(), IIntIds)

        >>> request.form = {'SEARCH_GROUP': 'b'}
        >>> widget.filter(items)
        [{'id': 1}, {'id': 2}]

        >>> request.form = {'SEARCH_GROUP': 'b',
        ...                 'SEARCH_TITLE': 'bet'}
        >>> widget.filter(items)
        [{'id': 1}]

   The search is case insensitive:

        >>> request.form = {'SEARCH_TITLE': 'AlphA'}
        >>> widget.filter(items)
        [{'id': 0}]

   The search also searches through usernames:

        >>> request.form = {'SEARCH_TITLE': '1234'}
        >>> widget.filter(items)
        [{'id': 0}]

    If clear search button is clicked, the form attribute is cleared,
    and all items are displayed:

        >>> request.form['CLEAR_SEARCH'] = 'Yes'

        >>> widget.filter(items)
        [{'id': 0}, {'id': 1}, {'id': 2}]
        >>> request.form['SEARCH_TITLE']
        ''

        >>> request.form['SEARCH_GROUP']
        ''

    """

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
        setUp=setUp, tearDown=tearDown,
        optionflags=doctest.ELLIPSIS|doctest.REPORT_NDIFF))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Views for the sample data generation
"""
from zope.publisher.browser import BrowserView
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import queryMultiAdapter, getUtilitiesFor
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.sampledata.generator import generate

from interfaces import ISampleDataPlugin

class SampleDataView(BrowserView):

    seed = 'SchoolTool'
    __used_for__ = ISchoolToolApplication

    template = ViewPageTemplateFile("sampledata.pt")

    work_done = property(lambda self: hasattr(self, 'times'))

    def __call__(self):
        self.update()
        return self.template()

    def update(self):
        if 'seed' in self.request:
            self.seed = self.request['seed']
            if self.seed == '':
                self.seed = None
        if 'CANCEL' in self.request:
            self.request.response.redirect(
                absoluteURL(self.context, self.request))
        if 'SUBMIT' in self.request:
            # TODO: maybe clear database here
            self.times = generate(self.context, self.seed,
                                  pluginNames=self._getSelectedPlugins())

    def _getSelectedPlugins(self):
        prefix = 'plugin.'
        for key in self.request.keys():
            if key.startswith(prefix):
                yield key[len(prefix):]

    def getPlugins(self):
        selectedPlugins = self._getSelectedPlugins()
        times = generate(self.context, self.seed,
                         dry_run=True, pluginNames=selectedPlugins)
        plugins = [obj for name, obj in getUtilitiesFor(ISampleDataPlugin)]
        result = []
        if 'CLEAR' in self.request:
            return plugins
        for plugin in plugins:
            data = {'name':plugin.name}
            if plugin.name in times.keys():
                data['selected'] = True
                view = queryMultiAdapter((plugin, self.request),
                                         name="options")
                data['view'] = view
            result.append(data)
        return result



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Interfaces for SchoolTool calendar browser views.
"""

from zope.viewlet.interfaces import IViewletManager
from zope.interface import Interface

from schooltool.calendar.browser.interfaces import IEventForDisplay
from schooltool.calendar.browser.interfaces import IHaveEventLegend


class ICalendarProvider(Interface):
    """Calendar provider.

    Subscription adapters providing this interface will be used to
    combine the list of calendars that will be displayed.
    """

    def getCalendars():
        """Gets a list of calendars to display.

        Yields tuples (calendar, color1, color2).
        """


class ICalendarMenuViewlet(Interface):
    """Marker interface so we could use custom crowd for View Calendar menu item"""


class IManageMenuViewletManager(IViewletManager):
    """Provides a viewlet hook for the management menu items."""


class ISchoolMenuViewletManager(IViewletManager):
    """Provides a viewlet hook for the school menu items."""


class IReportPageTemplate(Interface):

    def stylesheet():
        """Render the stylesheet contents."""

    def __call__():
        """Render the page template itself."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for basic person vocabularies.
"""
import unittest
import doctest

from schooltool.group.interfaces import IGroupContainer
from zope.app.testing import setup
from zope.component import provideAdapter
from zope.interface import implements


def doctest_GroupVocabulary():
    """Tests for GroupVocabulary

    If the context of a vocabulary is not a person, all groups from the
    group container are returned:

        >>> from schooltool.group.interfaces import IGroup
        >>> class GroupStub(object):
        ...     implements(IGroup)
        ...     def __init__(self, title):
        ...         self.title = title
        ...         self.__name__ = title.lower()

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> class STAppStub(dict):
        ...     implements(ISchoolToolApplication)
        ...     def __init__(self, context):
        ...         self['groups'] = {'a': GroupStub('A'),
        ...                           'b': GroupStub('B'),
        ...                           'some-group': GroupStub('Some-Group')}

        >>> provideAdapter(STAppStub, adapts=[None])
        >>> provideAdapter(lambda app: app['groups'],
        ...                adapts=[ISchoolToolApplication],
        ...                provides=IGroupContainer)

        >>> from schooltool.basicperson.vocabularies import GroupVocabulary
        >>> vocabulary = GroupVocabulary(None)
        >>> [group.token for group in vocabulary]
        ['a', 'b', 'some-group']

    If the context of the vocabulary is a person - list all the groups
    he belongs to as possible grade classes:

        >>> class SectionStub(object):
        ...     def __init__(self, name):
        ...         self.__name__ = name

        >>> from schooltool.person.interfaces import IPerson
        >>> class PersonStub(object):
        ...     implements(IPerson)
        ...     def __init__(self):
        ...         self.groups = map(GroupStub, ['A', 'B'])
        ...         self.groups.extend(map(SectionStub, ['s1', 's2']))

        >>> person = PersonStub()
        >>> vocabulary = GroupVocabulary(person)
        >>> [group.token for group in vocabulary]
        ['a', 'b']

    """


def doctest_AdvisorVocabulary():
    """Tests for AdvisorVocabulary

    If the context of a vocabulary is not a person, all groups from the
    group container are returned:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> class GroupStub(object):
        ...     def __init__(self, members):
        ...         self.members = members

        >>> class PersonStub(object):
        ...     def __init__(self, name):
        ...         self.__name__ = self.title = name

        >>> class STAppStub(dict):
        ...     implements(ISchoolToolApplication)
        ...     def __init__(self, context):
        ...         self['groups'] = {}
        ...         members = map(PersonStub, ['john', 'sarrah'])
        ...         self['groups']['teachers'] = GroupStub(members)

        >>> provideAdapter(STAppStub, adapts=[None])
        >>> provideAdapter(lambda app: app['groups'],
        ...                adapts=[ISchoolToolApplication],
        ...                provides=IGroupContainer)

        >>> from schooltool.basicperson.vocabularies import AdvisorVocabulary
        >>> vocabulary = AdvisorVocabulary(None)
        >>> [person.value.__name__ for person in vocabulary]
        ['john', 'sarrah']

    """


def setUp(test):
    setup.placelessSetUp()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return doctest.DocTestSuite(optionflags=optionflags,
                                setUp=setUp, tearDown=tearDown)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.relationship

This module contains a `setUpRelationships` function that can be used in
unit test setup code to register an IRelationshipLinks adapter for IAnnotatable
objects.  There are also `setUp` and `tearDown` functions that perform the
necessary Zope 3 placeless setup that is needed to make annotations and
relationships work for IAttributeAnnotatable objects.

This module also contains some stub objects for use in tests (SomeObject and
SomeContained).
"""

from zope.app.testing import setup
from zope.interface import implements
from zope.location.pickling import LocationCopyHook
from zope.component import provideAdapter
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.contained import Contained


class SomeObject(object):
    """A simple annotatable object for tests."""

    implements(IAttributeAnnotatable)

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return self._name


class SomeContained(SomeObject, Contained):
    """A simple annotatable contained object for tests."""


class URIStub(object):
    def __init__(self, uri):
        self.uri = uri
    def __cmp__(self, other):
        return cmp(self.uri, other.uri)


def setUp(test=None):
    """Set up for schooltool.relationship doctests.

    Calls Zope's placelessSetUp, sets up annotations and relationships.
    """
    setup.placelessSetUp()
    setup.setUpAnnotations()
    setUpRelationships()


def tearDown(test=None):
    """Tear down for schooltool.relationshp doctests."""
    setup.placelessTearDown()


def setUpRelationships():
    """Set up the adapter from IAnnotatable to IRelationshipLinks.

    This function is created for use in unit tests.  You should call
    zope.app.testing.setup.placelessSetUp before calling this function
    (and don't forget to call zope.app.testing.setup.placelessTearDown after
    you're done).  You should also call zope.app.testing.setup.setUpAnnotations
    to get a complete test fixture.
    """
    from zope.annotation.interfaces import IAnnotatable
    from zope.component import provideAdapter
    from schooltool.relationship.interfaces import IRelationshipLinks
    from schooltool.relationship.annotatable import getRelationshipLinks
    provideAdapter(getRelationshipLinks, (IAnnotatable,), IRelationshipLinks)
    provideAdapter(LocationCopyHook)





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for catalogs
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements, Interface
from zope.interface.verify import verifyObject
from zope.component import provideAdapter
from zope.component.hooks import getSite
from zope.site import SiteManagerContainer
from zope.site.folder import rootFolder

from schooltool.testing.setup import ZCMLWrapper
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ICatalogs
from schooltool.app.catalog import getAppCatalogs


class AppStub(dict, SiteManagerContainer):
    implements(ISchoolToolApplication)


class CatalogStub(dict):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.name)


def provideApplicationStub():
    app = AppStub()
    provideAdapter(
        lambda ignored: app,
        adapts=(None,),
        provides=ISchoolToolApplication)
    return app


def doctest_Catalogs():
    """Tests for Catalogs.

        >>> app = provideApplicationStub()

    There is an adapter that sets up and returns the catalogs container for the app.

        >>> list(app.items())
        []

        >>> catalogs = ICatalogs(app)

        >>> list(app.items())
        [('schooltool.app.catalog:Catalogs',
          <schooltool.app.catalog.Catalogs object at ...>)]

        >>> verifyObject(ICatalogs, catalogs)
        True

        >>> catalogs is app.values()[0]
        True

    Catalogs is a dict-like container.

        >>> print dict(catalogs)
        {}

    """


def doctest_PrepareCatalogContainer():
    """Tests for PrepareCatalogContainer.

        >>> class VersionedCatalogStub(object):
        ...     def __init__(self, expired=False):
        ...         self.expired = expired
        ...     def __repr__(self):
        ...         return '<Stub expired=%s>' % self.expired

    Let's build an app with some catalog entries.

        >>> app = provideApplicationStub()
        >>> catalogs = ICatalogs(app)

        >>> catalogs['cat-1'] = VersionedCatalogStub(expired=False)
        >>> catalogs['cat-2'] = VersionedCatalogStub(expired=True)
        >>> catalogs['cat-3'] = VersionedCatalogStub(expired=False)

    PrepareCatalogContainer is supposed to be executed as the first action
    of application catalog startup.

        >>> from schooltool.app.interfaces import ICatalogStartUp
        >>> from schooltool.app.catalog import PrepareCatalogContainer

        >>> action = PrepareCatalogContainer(app)
        >>> verifyObject(ICatalogStartUp, action)
        True

    When executed it marks all catalogs expired.

        >>> action()

        >>> print sorted(ICatalogs(app).items())
        [(u'cat-1', <Stub expired=True>),
         (u'cat-2', <Stub expired=True>),
         (u'cat-3', <Stub expired=True>)]

    """


def doctest_ExpiredCatalogCleanup():
    """Tests for ExpiredCatalogCleanup.

        >>> class VersionedCatalogStub(object):
        ...     def __init__(self, expired=False):
        ...         self.expired = expired
        ...     def __repr__(self):
        ...         return '<Stub expired=%s>' % self.expired

    Let's build an app with some catalog entries.

        >>> app = provideApplicationStub()
        >>> catalogs = ICatalogs(app)

        >>> catalogs['cat-1'] = VersionedCatalogStub(expired=False)
        >>> catalogs['cat-2'] = VersionedCatalogStub(expired=True)
        >>> catalogs['cat-3'] = VersionedCatalogStub(expired=False)

    ExpiredCatalogCleanup is supposed to be executed as the last action
    of application catalog startup.

        >>> from schooltool.app.interfaces import ICatalogStartUp
        >>> from schooltool.app.catalog import ExpiredCatalogCleanup

        >>> action = ExpiredCatalogCleanup(app)
        >>> verifyObject(ICatalogStartUp, action)
        True

    When executed it removes expired catalogs.

        >>> action()

        >>> print sorted(ICatalogs(app).items())
        [(u'cat-1', <Stub expired=False>),
         (u'cat-3', <Stub expired=False>)]

    """


def doctest_CatalogFactory():
    """Tests for CatalogFactory.

        >>> app = provideApplicationStub()

    CatalogFactory is a base class for creating and registering versioned catalogs.

        >>> from schooltool.app.catalog import CatalogFactory
        >>> factory = CatalogFactory(app)

        >>> factory.createCatalog()
        Traceback (most recent call last):
        ...
        NotImplementedError

        >>> factory.setIndexes(None)
        Traceback (most recent call last):
        ...
        NotImplementedError

    It is also an action to be executed on application catalog startup.

        >>> from schooltool.app.interfaces import ICatalogStartUp

        >>> class CatalogFactoryForTest(CatalogFactory):
        ...    version = 1
        ...    def createCatalog(self):
        ...        return CatalogStub('test')
        ...    def setIndexes(self, catalog):
        ...        catalog['idx-1'] = 'Stub index 1'

        >>> factory = CatalogFactoryForTest(app)

        >>> verifyObject(ICatalogStartUp, factory)
        True

    It defines some control on when it wants to be executed.

        >>> factory.after
        ('prepare-catalog-container',)

        >>> factory.before
        ('expired-catalog-cleanup',)

    The class itself has some helpers to fetch the catalog from app.

        >>> print CatalogFactoryForTest.key()
        catalog:schooltool.app.tests.test_catalog.CatalogFactoryForTest

        >>> print CatalogFactoryForTest.get()
        None

    Well, the catalog was not set up yet.  Let's build it.

        >>> from schooltool.app.interfaces import IVersionedCatalog

        >>> factory()

        >>> catalog = CatalogFactoryForTest.get()

        >>> print catalog
        <CatalogStub 'test'>

        >>> sorted(catalog.items())
        [('idx-1', 'Stub index 1')]

    Both catalog and it's version entry are stored in app's catalog container.

        >>> version_entry = catalog.__parent__

        >>> print version_entry
        <VersionedCatalog v. u'1'>: <CatalogStub 'test'>

        >>> verifyObject(IVersionedCatalog, version_entry)
        True

        >>> print version_entry.__parent__
        <schooltool.app.catalog.Catalogs object at ...>

        >>> version_entry.__parent__ is ICatalogs(app)
        True

        >>> print version_entry.__name__
        catalog:schooltool.app.tests.test_catalog.CatalogFactoryForTest

        >>> print version_entry.__name__ == CatalogFactoryForTest.key()
        True

    """


def doctest_CatalogFactory_versioning():
    """Tests for CatalogFactory handling of catalog versions.

        >>> class CatalogCreatorMixin(object):
        ...    def createCatalog(self):
        ...        return CatalogStub('test-1')
        ...    def setIndexes(self, catalog):
        ...        catalog['idx-1'] = 'Stub index 1'

        >>> app = provideApplicationStub()
        >>> catalogs = ICatalogs(app)

    Let's create a catalog first.

        >>> from schooltool.app.catalog import CatalogFactory
        >>> class Factory1(CatalogCreatorMixin, CatalogFactory):
        ...    version = 1

        >>> factory1 = Factory1(app)

        >>> factory1()
        >>> catalog1 = Factory1.get()

        >>> print catalog1
        <CatalogStub 'test-1'>

        >>> print sorted(catalogs.keys())
        [u'catalog:schooltool.app.tests.test_catalog.Factory1']

    If the version stays the same, factory leaves the catalog untouched, but
    changes it's expired status.

        >>> class Factory1(CatalogCreatorMixin, CatalogFactory):
        ...    version = 1

        >>> catalog1.__parent__.expired = True

        >>> factory1 = Factory1(app)
        >>> factory1()

        >>> Factory1.get() is catalog1
        True

        >>> catalog1.__parent__.expired
        False

        >>> print sorted(catalogs.keys())
        [u'catalog:schooltool.app.tests.test_catalog.Factory1']

    If we changed the version, catalog gets re-created.

        >>> class Factory1(CatalogCreatorMixin, CatalogFactory):
        ...    version = 1.2
        ...    def setIndexes(self, catalog):
        ...        catalog['idx-1'] = 'Stub index 1 point 2'

        >>> factory12 = Factory1(app)
        >>> factory12()

        >>> catalog12 = Factory1.get()
        >>> catalog12 is catalog1
        False

        >>> print sorted(catalog12.items())
        [('idx-1', 'Stub index 1 point 2')]

    CatalogFactory actually uses the getVersion method, not the version
    attribute.  Let's override getVersion

        >>> class Factory1(CatalogCreatorMixin, CatalogFactory):
        ...    version = 1.2
        ...    def getVersion(self):
        ...        return 'stable-version'

        >>> factory = Factory1(app)
        >>> factory()

        >>> catalog_stable = Factory1.get()

        >>> catalog_stable is catalog12
        False

        >>> print Factory1.get().__parent__
        <VersionedCatalog v. 'stable-version'>: <CatalogStub 'test-1'>

        >>> class Factory1(CatalogCreatorMixin, CatalogFactory):
        ...    version = 1.314
        ...    def getVersion(self):
        ...        return 'stable-version'

        >>> factory1 = Factory1(app)
        >>> factory1()

        >>> Factory1.get() is catalog_stable
        True

    """


def doctest_CatalogImplementing():
    """Tests for CatalogImplementing.  This is a factory of catalogs that
    contain only objects implementing given interface.

    Say we have two objects implementing very similar interfaces.

        >>> from zope.schema import TextLine

        >>> class TitledObject(object):
        ...    def __init__(self, title):
        ...        self.title = title
        ...    def __repr__(self):
        ...        return '<%s (%r)>' % (self.__class__.__name__, self.title)

        >>> class IFoo(Interface):
        ...     title = TextLine(title=u"Title")
        >>> class Foo(TitledObject):
        ...     implements(IFoo)

        >>> class IBar(Interface):
        ...     title = TextLine(title=u"Title")
        >>> class Bar(TitledObject):
        ...     implements(IBar)

        >>> app = provideApplicationStub()
        >>> app['objects'] = {
        ...     1: Foo('one'),
        ...     2: Bar('two'),
        ...     3: Foo('three'),
        ...     }

    Let's create a catalog that indexes titles of IFoo.

        >>> from zope.container.contained import Contained
        >>> from schooltool.app.catalog import CatalogImplementing

        >>> class TitleIndex(Contained):
        ...     def __init__(self):
        ...         self.data = []
        ...     def index_doc(self, id, value):
        ...         self.data.append((id, value))

        >>> class CatalogTitles(CatalogImplementing):
        ...    version = 1
        ...    interface = IFoo
        ...    def setIndexes(self, catalog):
        ...        catalog['title'] = TitleIndex()

        >>> factory = CatalogTitles(app)
        >>> factory()

        >>> catalog = CatalogTitles.get()
        >>> print catalog.__parent__
        <VersionedCatalog v.
            u'interface:schooltool.app.tests.test_catalog.IFoo,
              version:1'>:
        <zc.catalog.extentcatalog.Catalog object at ...>

    This catalog only indexes objects implementing IFoo.

        >>> def index_app(catalog):
        ...     for docid, doc in app['objects'].items():
        ...         catalog.index_doc(docid, doc)
        ...     print sorted(catalog['title'].data)

        >>> index_app(CatalogTitles.get())
        [(1, <Foo ('one')>), (3, <Foo ('three')>)]

    Catalog indexes are untouched after app restart.

        >>> factory = CatalogTitles(app)
        >>> factory()

        >>> CatalogTitles.get() is catalog
        True

        >>> print sorted(CatalogTitles.get()['title'].data)
        [(1, <Foo ('one')>), (3, <Foo ('three')>)]

    If we change the version, catalog will be re-created, as expected.

        >>> class CatalogTitles(CatalogImplementing):
        ...    version = 2
        ...    interface = IFoo
        ...    def setIndexes(self, catalog):
        ...        catalog['title'] = TitleIndex()

        >>> factory = CatalogTitles(app)
        >>> factory()

        >>> catalog2 = CatalogTitles.get()
        >>> catalog2 is catalog
        False

    We trust other machinery to reindex the catalog.

        >>> print sorted(CatalogTitles.get()['title'].data)
        []

    If we change the interface, catalog also gets re-created.

        >>> index_app(CatalogTitles.get())
        [(1, <Foo ('one')>), (3, <Foo ('three')>)]

        >>> class CatalogTitles(CatalogImplementing):
        ...    version = 2
        ...    interface = IBar
        ...    def setIndexes(self, catalog):
        ...        catalog['title'] = TitleIndex()

        >>> factory = CatalogTitles(app)
        >>> factory()

        >>> CatalogTitles.get() is catalog2
        False

        >>> index_app(CatalogTitles.get())
        [(2, <Bar ('two')>)]

        >>> catalog3 = CatalogTitles.get()
        >>> print catalog3.__parent__
        <VersionedCatalog v.
            u'interface:schooltool.app.tests.test_catalog.IBar,
              version:2'>:
        <zc.catalog.extentcatalog.Catalog object at ...>

    """


def doctest_AttributeCatalog():
    """Tests for AttributeCatalog.  This is a factory of catalogs that
    index attributes of objects implementing given interface.

    Say we have an object we want to catalog.

        >>> from zope.schema import TextLine

        >>> class ITriplet(Interface):
        ...     a = TextLine(title=u"Title")
        ...     b = TextLine(title=u"Title")
        ...     c = TextLine(title=u"Title")
        >>> class Triplet(object):
        ...     implements(ITriplet)
        ...     def __init__(self, a, b, c):
        ...         self.a, self.b, self.c = a, b, c
        ...     def __repr__(self):
        ...         return '<%s (%s:%s:%s)>' % (
        ...             self.__class__.__name__,
        ...             self.a, self.b, self.c)

        >>> class Foo(object):
        ...     implements(Interface)
        ...     a = b = c = 0

        >>> app = provideApplicationStub()
        >>> app['objects'] = {
        ...     0: Foo(),
        ...     1: Triplet('one', 'eins', 'I'),
        ...     2: Triplet('two', 'zwei', 'II'),
        ...     3: Triplet('three', 'drei', 'III'),
        ...     }

    Let's create a catalog that indexes few attributes of the triplet.

        >>> from schooltool.app.catalog import AttributeCatalog

        >>> class CatalogTriplets(AttributeCatalog):
        ...    version = 1
        ...    interface = ITriplet
        ...    attributes = ('a', 'c')

        >>> factory = CatalogTriplets(app)
        >>> factory()

    And index some objects.

        >>> def index_app(catalog):
        ...     for docid, doc in app['objects'].items():
        ...         catalog.index_doc(docid, doc)
        ...     for name, index in catalog.items():
        ...         print 'catalog[%r]: %s' % (
        ...             name,
        ...             sorted(index.documents_to_values.items()))

        >>> index_app(CatalogTriplets.get())
        catalog[u'a']: [(1, 'one'), (2, 'two'), (3, 'three')]
        catalog[u'c']: [(1, 'I'), (2, 'II'), (3, 'III')]

    We can see that version of the catalog also includes the list of attributes,
    so the catalog will be recreated when attributes, interface or version changes.

        >>> catalog = CatalogTriplets.get()
        >>> print catalog.__parent__
        <VersionedCatalog v.
            u"attributes:('a', 'c'),
              interface:schooltool.app.tests.test_catalog.ITriplet,
              version:1">:
        <zc.catalog.extentcatalog.Catalog object at ...>

        >>> class CatalogTriplets(AttributeCatalog):
        ...    version = 1
        ...    interface = ITriplet
        ...    attributes = ('b')

        >>> factory = CatalogTriplets(app)
        >>> factory()

        >>> CatalogTriplets.get() is catalog
        False

        >>> index_app(CatalogTriplets.get())
        catalog[u'b']: [(1, 'eins'), (2, 'zwei'), (3, 'drei')]

    """


def doctest_catalog_subscribers():
    """Tests for catalog subscribers.

    Let's provide our subscribers.

        >>> from zope.component import provideHandler
        >>> from schooltool.app.catalog import indexDocSubscriber
        >>> from schooltool.app.catalog import reindexDocSubscriber
        >>> from schooltool.app.catalog import unindexDocSubscriber

        >>> provideHandler(indexDocSubscriber)
        >>> provideHandler(reindexDocSubscriber)
        >>> provideHandler(unindexDocSubscriber)

    And set up the event firing helpers.

        >>> from zope.component import provideUtility, queryUtility
        >>> from zope.event import notify
        >>> from zope.intid.interfaces import (
        ...     IIntIds, IntIdAddedEvent, IntIdRemovedEvent)
        >>> from zope.lifecycleevent import (
        ...     ObjectAddedEvent, ObjectRemovedEvent, ObjectModifiedEvent)

        >>> def addAndNotify(obj, obj_id):
        ...     util = queryUtility(IIntIds)
        ...     if util is None:
        ...         print "No IIntIds utility, so don't fire IntIdAddedEvent"
        ...         return
        ...     util[obj] = obj_id
        ...     print 'firing IntIdAddedEvent'
        ...     notify(IntIdAddedEvent(obj, ObjectAddedEvent(obj)))

        >>> def notifyModified(obj):
        ...     print 'firing ObjectModifiedEvent'
        ...     notify(ObjectModifiedEvent(obj))

        >>> def notifyAndRemove(obj):
        ...     util = queryUtility(IIntIds)
        ...     if util is None:
        ...         print "No IIntIds utility, so don't fire IntIdRemovedEvent"
        ...         return
        ...     print 'firing IntIdRemovedEvent'
        ...     notify(IntIdRemovedEvent(obj, ObjectRemovedEvent(obj)))
        ...     del util[obj]

    When database is being set up, we may have no SchoolToolApplication and
    no IntIds utility.

        >>> class TestObj(object):
        ...     __parent__ = __name__ = None
        ...     def __init__(self, name):
        ...         self.name = name
        ...     def __repr__(self):
        ...         return '<%s %r>' % (self.__class__.__name__, self.name)

        >>> test_one = TestObj('missing_one')
        >>> addAndNotify(test_one, 1)
        No IIntIds utility, so don't fire IntIdAddedEvent

        >>> notifyModified(test_one)
        firing ObjectModifiedEvent

        >>> notifyAndRemove(test_one)
        No IIntIds utility, so don't fire IntIdRemovedEvent

    Having IntIds utility is not enough as catalogs live within the application
    itself.  Subscribers handle this case also.

        >>> class IntIdsStub(dict):
        ...     def getId(self, obj):
        ...         return self[obj]
        ...     def queryId(self, obj):
        ...         return self.get(obj)

        >>> provideUtility(IntIdsStub(), IIntIds)

        >>> test_two = TestObj('two')
        >>> addAndNotify(test_two, 2)
        firing IntIdAddedEvent

        >>> notifyModified(test_two)
        firing ObjectModifiedEvent

        >>> notifyAndRemove(test_two)
        firing IntIdRemovedEvent

    Let's provide an application and set up a catalog.

        >>> from schooltool.app.catalog import VersionedCatalog

        >>> class PrintingCatalogStub(CatalogStub):
        ...      def index_doc(self, doc_id, doc):
        ...          print 'CatalogStub(%r) indexed doc %s (%s)' % (
        ...              self.name, doc_id, doc)
        ...      def unindex_doc(self, doc_id):
        ...          print 'CatalogStub(%r) unindexed doc %s' % (
        ...              self.name, doc_id)

        >>> app = provideApplicationStub()
        >>> catalogs = ICatalogs(app)
        >>> catalogs['demo'] = VersionedCatalog(PrintingCatalogStub('demo'), 'v1')

    We can now see objects being indexed.

        >>> test_three = TestObj('three')
        >>> addAndNotify(test_three, 3)
        firing IntIdAddedEvent
        CatalogStub('demo') indexed doc 3 (<TestObj 'three'>)

        >>> test_three.name='three and a half'
        >>> notifyModified(test_three)
        firing ObjectModifiedEvent
        CatalogStub('demo') indexed doc 3 (<TestObj 'three and a half'>)

        >>> notifyAndRemove(test_three)
        firing IntIdRemovedEvent
        CatalogStub('demo') unindexed doc 3

    """


def setUp(test):
    setup.placefulSetUp()
    provideAdapter(getAppCatalogs)


def tearDown(test):
    setup.placefulTearDown()


def provideStubAdapter(factory, adapts=None, provides=None, name=u''):
    sm = getSite().getSiteManager()
    sm.registerAdapter(factory, required=adapts, provided=provides, name=name)


def unregisterStubAdapter(factory, adapts=None, provides=None, name=u''):
    sm = getSite().getSiteManager()
    sm.unregisterAdapter(factory, required=adapts, provided=provides, name=name)


def setUpIntegration(test):
    setup.placefulSetUp()
    zcml = ZCMLWrapper()
    zcml.setUp(
        namespaces={"": "http://namespaces.zope.org/zope"},
        i18n_domain='schooltool')
    zcml.include('schooltool.common', file='zcmlfiles.zcml')
    zcml.include('schooltool.app', file='catalog.zcml')
    root = rootFolder()
    root['app'] = provideApplicationStub()
    setup.createSiteManager(root['app'], setsite=True)
    test.globs.update({
        'zcml': zcml,
        'CatalogStub': CatalogStub,
        'provideAdapter': provideStubAdapter,
        'unregisterAdapter': unregisterStubAdapter,
        })


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=optionflags),
        doctest.DocFileSuite(
            'catalog-integration.txt',
            setUp=setUpIntegration, tearDown=tearDown,
            optionflags=optionflags),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
